# -*- coding: utf-8 -*-
"""
test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import pytest
import sys
import  os.path as path
import numpy as np
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import warnings

warnings.filterwarnings("ignore")

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../../"))
sys.path.insert(0, pth)
from Stoner import Data, __home__, Options
from Stoner.Core import typeHintedDict
from Stoner.plot.core import hsl2rgb
from Stoner.Image import ImageFile

from Stoner.plot.formats import DefaultPlotStyle

datadir = path.join(pth, "sample-data")

selfd = Data(path.join(__home__, "..", "sample-data", "New-XRay-Data.dql"))


warnings.filterwarnings("error")

def test_set_no_figs():
    global selfd
    assert Options.no_figs, "Default setting for no_figs option is incorrect."
    Options.no_figs = True
    e = selfd.clone
    ret = e.plot()
    assert ret is None, "Output of Data.plot() was not None when no_figs is True  and showfig is not set({})".format(
        type(ret)
    )
    Options.no_figs = False
    e.showfig = False
    ret = e.plot()
    assert isinstance(ret, Data), "Return value of Data.plot() was not self when Data.showfig=False ({})".format(
        type(ret)
    )
    e.showfig = True
    ret = e.plot()
    assert isinstance(ret, Figure), "Return value of Data.plot() was not Figure when Data.showfig=False({})".format(
        type(ret)
    )
    e.showfig = None
    ret = e.plot()
    assert ret is None, "Output of Data.plot() was not None when Data.showfig is None ({})".format(type(ret))
    Options.no_figs = True
    assert Options.no_figs, "set_option no_figs failed."
    selfd = Data(path.join(__home__, "..", "sample-data", "New-XRay-Data.dql"))
    selfd.showfig = False
    ret = selfd.plot()
    assert ret is None, "Output of Data.plot() was not None when no_figs is True ({})".format(type(ret))
    Options.no_figs = True
    plt.close("all")


@pytest.mark.filterwarnings("ignore:.*:matplotlib.MatplotlibDeprecationWarning")
def test_template_settings():
    template = DefaultPlotStyle(font__weight="bold")
    assert template["font.weight"] == "bold", "Setting ytemplate parameter in init failed."
    template(font__weight="normal")
    assert template["font.weight"] == "normal", "Setting ytemplate parameter in call failed."
    template["font.weight"] = "bold"
    assert template["font.weight"] == "bold", "Setting ytemplate parameter in setitem failed."
    del template["font.weight"]
    assert template["font.weight"] == "normal", "Resettting template parameter failed."
    keys = sorted([x for x in template])
    assert sorted(template.keys()) == keys, "template.keys() and template.iter() disagree."
    attrs = [x for x in dir(template) if template._allowed_attr(x)]
    length = len(dict(plt.rcParams)) + len(attrs)
    assert len(template) == length, "templa length wrong."


def test_plot_magic():
    selfd.figure()
    dpi = selfd.fig_dpi
    selfd.fig_dpi = dpi * 2
    assert selfd.fig.dpi == dpi * 2, "Failed to get/set attributes on current figure"
    vis = selfd.fig_visible
    selfd.fig_visible = not vis
    assert not selfd.fig_visible, "Setting/Getting figure.visible failed"
    plt.close("all")
    plt.figure()
    fn = plt.get_fignums()[0]
    selfd.fig = fn
    selfd.plot()
    assert len(plt.get_fignums()) == 1, "Setting Data.fig by integer failed."
    plt.close("all")
    selfd.plot(plotter=plt.semilogy)
    assert selfd.ax_lines[0].get_c() == "k", "Auto formatting of plot failed"
    selfd.plot(figure=False)
    selfd.plot(figure=1)
    assert len(plt.get_fignums()) == 2, "Plotting setting figure failed"
    assert len(selfd.ax_lines) == 2, "Plotting setting figure failed"
    selfd.figure(2)
    selfd.plot()
    selfd.plot(figure=True)
    assert len(plt.get_fignums()) == 2, "Plotting setting figure failed"
    assert len(selfd.ax_lines) == 3, "Plotting setting figure failed"
    plt.close("all")


def test_extra_plots():
    x = np.random.uniform(-np.pi, np.pi, size=5001)
    y = np.random.uniform(-np.pi, np.pi, size=5001)
    z = (np.cos(4 * np.sqrt(x**2 + y**2)) * np.exp(-np.sqrt(x**2 + y**2) / 3.0)) ** 2
    selfd2 = Data(x, y, z, column_headers=["X", "Y", "Z"], setas="xyz")
    selfd2.contour_xyz(projection="2d")  #
    assert len(plt.get_fignums()) == 1, "Setting Data.fig by integer failed."
    plt.close("all")
    X, Y, Z = selfd2.griddata(xlim=(-np.pi, np.pi), ylim=(-np.pi, np.pi))
    plt.imshow(Z)
    assert len(plt.get_fignums()) == 1, "Setting Data.fig by integer failed."
    plt.imshow(Z)
    plt.close("all")
    x, y = np.meshgrid(np.linspace(-np.pi, np.pi, 10), np.linspace(-np.pi, np.pi, 10))
    z = np.zeros_like(x)
    w = np.cos(np.sqrt(x**2 + y**2))
    q = np.arctan2(x, y)
    u = np.abs(w) * np.cos(q)
    v = np.abs(w) * np.sin(q)
    selfd3 = Data(x.ravel(), y.ravel(), z.ravel(), u.ravel(), v.ravel(), w.ravel(), setas="xyzuvw")
    selfd3.plot()
    assert len(plt.get_fignums()) == 1, "Setting Data.fig by integer failed."
    plt.close("all")
    # i=ImageFile(path.join(__home__,"..","sample-data","Kermit.png"))
    # selfd3=Data(i)
    # selfd3.data=i.data
    # selfd3.plot_matrix()


def test_misc_funcs():
    assert np.all(hsl2rgb(0.5, 0.5, 0.5) == np.array([[63, 191, 191]]))
    selfd.load(selfd.filename, Q=True)
    selfd.plot()
    selfd.x2()
    selfd.setas = ".yx"
    selfd.plot()
    assert len(selfd.fig_axes) == 2, "Creating a second X axis failed"
    plt.close("all")
    for i in range(4):
        selfd.subplot(2, 2, i + 1)
        selfd.plot()
    assert len(selfd.fig_axes) == 4, "Creating subplots failed"
    selfd.close()


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Provide a mixin class the provides sequence and mapping like interfaces for Data."""

__all__ = ["DataFileInterfacesMixin"]

import numpy as np

from ..compat import string_types, _pattern_type
from ..tools import isiterable


class DataFileInterfacesMixin:
    """Implement the required methods for a sequence and mapping type object."""

    def __contains__(self, item):
        """Operator function for membertship tests - used to check metadata contents.

        Args:
            item(string):
                name of metadata key

        Returns:
            (bool):
                True if item in self.metadata
        """
        return item in self.metadata

    def __delitem__(self, item):
        """Implement row or metadata deletion.

        Args:
            item (ingteger or string):
                row index or name of metadata to delete
        """
        if isinstance(item, string_types):
            del self.metadata[item]
        else:
            self.del_rows(item)

    def __getitem__(self, name):
        """Return either a row or iterm of metadata.

        Args:
            name (string or slice or int):
                The name, slice or number of the part of the
            :py:class:`DataFile` to be returned.

        Returns:
            mixed: an item of metadata or row(s) of data.

        -   If name is an integer then the corresponding single row will be returned
        -   if name is a slice, then the corresponding rows of data will be returned.
        -   If name is a string then the metadata dictionary item             with the correspondoing key will be
            returned.
        -   If name is a numpy array then the corresponding rows of the data are returned.
        -   If a tuple is supplied as the argument then there are a number of possible behaviours.
            -   If the first element of the tuple is a string, then it is assumed that it is the nth element of the
                named metadata is required.
            -   Otherwise itis assumed that it is a particular element within a column determined by the second
                part of the tuple that is required.

        Examples:
            DataFile['Temp',5] would return the 6th element of the
            list of elements in the metadata called 'Temp', while

            DataFile[5,'Temp'] would return the 6th row of the data column
            called 'Temp'

            and DataFile[5,3] would return the 6th element of the
            4th column.
        """
        if isinstance(name, string_types + (_pattern_type,)):
            try:
                ret = self.metadata[name]
            except KeyError:
                try:
                    ret = self.data[name]
                except KeyError as err:
                    raise KeyError(f"{name} was neither a key in the metadata nor a column in the main data.") from err
        elif isinstance(name, tuple) and isinstance(name[0], string_types):
            try:
                rest = name[1:]
                ret = self.metadata[name[0]]
                ret = ret.__getitem__(*rest)
            except KeyError:
                try:
                    ret = self.data[name]
                except KeyError as err:
                    raise KeyError(f"{name} was neither a key in the metadata nor a column in the main data.") from err
        elif isinstance(name, tuple) and name in self.metadata:
            ret = self.metadata[name]
        else:
            ret = self.data[name]
        return ret

    def __iter__(self):
        """Provide agenerator for iterating.

        Pass through to :py:meth:`DataFile.rows` for the actual work.

        Returns:
            Next row
        """
        for r in self.rows(False):
            yield r

    def __len__(self):
        """Return the length of the data.

        Returns: Returns the number of rows of data
        """
        if np.prod(self.data.shape) > 0:
            return np.shape(self.data)[0]
        return 0

    def __setitem__(self, name, value):
        """Handle direct setting of with metadata items or data array elements.

        Args:
            name (string, tuple):
                The string key used to access the metadata or a tuple index into data
            value (any):
                The value to be written into the metadata or data/

        Notes:
            If name is a string or already exists as key in metadata, this setitem will set metadata values,
            otherwise if name is a tuple then if the first elem,ent in a string it checks to see if that is an
            existing metadata item that is iterable, and if so, sets the metadta. In all other circumstances,
            it attempts to set an item in the main data array.
        """
        if isinstance(name, string_types) or str(name) in self.metadata:
            self.metadata[name] = value
        elif isinstance(name, tuple):
            if isinstance(name[0], string_types) and name[0] in self.metadata and isiterable(self.metadata[name[0]]):
                if len(name) == 2:
                    key = name[0]
                    name = name[1]
                else:
                    key = name[0]
                    name = tuple(name[1:])
                self.metadata[key][name] = value
            else:
                self.data[name] = value
        else:
            self.data[name] = value

    def count(self, value=None, axis=0, col=None):
        """Count the number of un-masked elements in the :py:class:`DataFile`.

        Keywords:
            valiue (float):
                Value to count for
            axis (int):
                Which axis to count the unmasked elements along.
            col (index, None):
                Restrict to counting in a specific column. If left None, then the current 'y' column is used.

        Returns:
            (int):
                Number of unmasked elements.
        """
        _ = self._col_args(ycol=col)
        if _.ycol is not None:
            tmp = self.column(_.ycol)
        else:
            tmp = self.data
        if value is not None:
            args = np.argwhere(tmp == value)
            return args.size
        return tmp.count(axis)

    def insert(self, index, obj):
        """Implement the insert method."""
        self.data = np.insert(self.data, index, obj, axis=0).view(type(self.data))



# -*- coding: utf-8 -*-
"""Provide the mixin classes for the DataFile Operators."""

__all__ = ["DataFileOperatorsMixin"]
import numpy as np

from ..compat import index_types, string_types
from ..tools import isiterable
from .utils import add_core, and_core, mod_core, sub_core
from . import _setas, DataArray


class DataFileOperatorsMixin:
    """Provides the operator mixins for DataFile like objects."""

    def __add__(self, other):
        """Implement a + operator to concatenate rows of data.

        Args:
            other (numpy arra `Stoner.Core.DataFile` or a dictionary or a list):
                The object to be added to the DataFile

        Note:
            -   if other is a dictionary then the keys of the dictionary are passed to
                :py:meth:`find_col` to see if they match a column, in which case the
                corresponding value will be used for theat column in the new row.
                Columns which do not have a matching key will be set to NaN. If other has keys
                that are not found as columns in self, additional columns are added.
            -   If other is a list, then the add method is called recursively for each element
                of the list.
            -   Returns: A Datafile object with the rows of @a other appended
                to the rows of the current object.
            -   If other is a 1D numopy array with the same number of
                elements as their are columns in @a self.data then the numpy
                array is treated as a new row of data If @a ither is a 2D numpy
                array then it is appended if it has the same number of
                columns and @a self.data.
        """
        newdata = self.clone
        return add_core(other, newdata)

    def __iadd__(self, other):
        """Implement a += operator to concatenate rows of data inplace.

        Args:
            other (numpy arra `Stoner.Core.DataFile` or a dictionary or a list):
                The object to be added to the DataFile

        Note:
            -   if other is a dictionary then the keys of the dictionary are passed to
                :py:meth:`find_col` to see if they match a column, in which case the
                corresponding value will be used for theat column in the new row.
                Columns which do not have a matching key will be set to NaN. If other has keys
                that are not found as columns in self, additional columns are added.
            -   If other is a list, then the add method is called recursively for each element
                of the list.
            -   Returns: A Datafile object with the rows of @a other appended
                to the rows of the current object.
            -   If other is a 1D numopy array with the same number of
                elements as their are columns in @a self.data then the numpy
                array is treated as a new row of data If @a ither is a 2D numpy
                array then it is appended if it has the same number of
                columns and @a self.data.
        """
        newdata = self
        return add_core(other, newdata)

    def __and__(self, other):
        """Implement the & operator to concatenate columns of data in a :py:class:`DataFile` object.

        Args:
            other  (numpy array or :py:class:`DataFile`):
                Data to be added to this DataFile instance

        Returns:
            ():py:class:`DataFile`):
                new Data object with the columns of other concatenated as new columns at the end of the self object.

        Note:
            Whether other is a numopy array of :py:class:`DataFile`, it must
            have the same or fewer rows than the self object.
            The size of @a other is increased with zeros for the extra rows.
            If other is a 1D numpy array it is treated as a column vector.
            The new columns are given blank column headers, but the
            length of the :py:meth:`column_headers` is
            increased to match the actual number of columns.
        """
        # Prep the final DataFile
        newdata = self.clone
        return and_core(other, newdata)

    def __iand__(self, other):
        """Implement the &= operator to concatenate columns of data in a :py:class:`DataFile` object.

        Args:
            other  (numpy array or :py:class:`DataFile`):
                Data to be added to this DataFile instance

        Returns:
            ():py:class:`DataFile`):
                new Data object with the columns of other concatenated as new columns at the end of the self object.

        Note:
            Whether other is a numopy array of :py:class:`DataFile`, it must
            have the same or fewer rows than the self object.
            The size of @a other is increased with zeros for the extra rows.
            If other is a 1D numpy array it is treated as a column vector.
            The new columns are given blank column headers, but the
            length of the :py:meth:`column_headers` is
            increased to match the actual number of columns.
        """
        newdata = self
        return and_core(other, newdata)

    def __eq__(self, other):
        """Equality operator.

        Args:
            other (DataFile):
                The object to test for equality against.

        Returns:
            bool:
                True if data, column headers and metadata are equal.
        """
        if not isinstance(other, DataFileOperatorsMixin):  # Start checking we're all DataFile's
            return False
        if self.data.shape != other.data.shape:  # Check we have the same data
            return False
        if not np.all(self.data[~np.isnan(self.data)] == other.data[~np.isnan(other.data)]):  # Don't compare nan!
            return False
        if len(self.column_headers) != len(other.column_headers) or np.any(
            [c1 != c2 for c1, c2 in zip(self.column_headers, other.column_headers)]
        ):  # And the same headers
            return False
        if not super().__eq__(other):  # Check metadata
            return False
        # Ok give up, we are the same!
        return True

    def __floordiv__(self, other):
        """Just aslias for self.column(other)."""
        if not isinstance(other, index_types):
            return NotImplemented
        return self.column(other)

    def __lshift__(self, other):
        """Convert a string or iterable to a new DataFile like object.

        Args:
            other (string or iterable object):
                Used to source the DataFile object

        Returns:
            (DataFile):
                A new :py:class:`DataFile` object

        Todo:
            Make code work better with streams

        Overird the left shift << operator for a string or an iterable object to import using the :py:meth:`__
        read_iterable` function.
        """
        newdata = type(self)()
        if isinstance(other, string_types):
            lines = map(lambda x: x, other.splitlines())
            newdata.__read_iterable(lines)
        elif isiterable(other):
            newdata.__read_iterable(other)
        else:
            return NotImplemented
        return type(self)(newdata)

    def __mod__(self, other):
        """Overload the % operator to mean column deletion.

        Args:
            Other (column index):
                column(s) to delete.

        Return:
            (self):
                A copy of self with a column deleted.
        """
        newdata = self.clone
        return mod_core(other, newdata)

    def __imod__(self, other):
        """Overload the % operator to mean in-place column deletion.

        Args:
            Other (column index):
                column(s) to delete.

        Return:
            (self):
                A copy of self with a column deleted.
        """
        newdata = self
        return mod_core(other, newdata)

    def __sub__(self, other):
        """Implement what to do when subtraction operator is used.

        Args:
            other (int,list of integers):
                Delete row(s) from data.

        Returns:
            (DataFile):
                A :py:data:`DataFile` with rows removed.
        """
        newdata = self.clone
        return sub_core(other, newdata)

    def __isub__(self, other):
        """Implement what to do when subtraction operator is used.

        Args:
            other (int,list of integers):
                Delete row(s) from data.

        Returns:
            (self):
                The :py:data:`DataFile` with rows removed.
        """
        newdata = self
        return sub_core(other, newdata)

    def __invert__(self):
        """Swap x and y column assignments around."""
        ret = self.clone
        setas = list(self.setas)
        cols = self.setas._cols
        if cols["axes"] == 2:
            swaps = zip(["ycol", "yerr"], ["x", "d"])
        elif cols["axes"] >= 3:
            swaps = zip(["ycol", "zcol", "yerr", "zerr"], ["z", "x", "f", "d"])
        setas[cols["xcol"]] = "y"
        if cols["has_xerr"]:
            setas[cols["xerr"]] = "e"
        for cname, nlet in swaps:
            for c in cols[cname]:
                setas[c] = nlet
        ret.setas = setas
        return ret  #

    def __read_iterable(self, reader):
        """Read a string representation of py:class:`DataFile` in line by line."""
        if isiterable(reader):
            reader = iter(reader)
        if "readline" in dir(reader):  # Filelike iterator
            readline = reader.readline
        elif "__next__" in dir(reader):  # Python v3 iterator
            readline = reader.__next__

        else:
            return NotImplemented
        row = readline().split("\t")
        if row[0].strip() == "TDI Format 1.5":
            fmt = 1.5
        elif row[0].strip() == "TDI Format=Text 1.0":
            fmt = 1.0
        else:
            raise RuntimeError("Not a TDI File")
        col_headers_tmp = [x.strip() for x in row[1:]]
        cols = len(col_headers_tmp)
        self._data._setas = _setas("." * cols)
        self.data = DataArray([], setas=self._data._setas)
        for r in reader:
            if r.strip() == "":  # Blank line
                continue
            row = r.rstrip().split("\t")
            cols = max(cols, len(row) - 1)
            if row[0].strip() != "":
                md = row[0].split("=")
                if len(md) >= 2:
                    md[1] = "=".join(md[1:])
                elif len(md) <= 1:
                    md.extend(["", ""])

                if fmt == 1.5:
                    self.metadata[md[0].strip()] = md[1].strip()
                elif fmt == 1.0:
                    self.metadata[md[0].strip()] = self.metadata.string_to_type(md[1].strip())
            if len(row) < 2:
                continue
            self.data = np.append(self.data, self._conv_float(row[1:]))
        self.data = np.reshape(self.data, (-1, cols))
        self.column_headers = [f"Column {i}" for i in range(cols)]
        for i, head_temp in enumerate(col_headers_tmp):
            self.column_headers[i] = head_temp
        self["TDI Format"] = fmt



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 19 12:04:58 2020

@author: phygbu
"""

import sys
import pytest
from matplotlib.backend_bases import Event
import threading
import time
import numpy as np
import Stoner

ret_pth = Stoner.__homepath__ / ".." / "sample-data" / "TDI_Format_RT.txt"

# Horrible hack to patch QFileDialog  for testing

import Stoner.tools.widgets as widgets
from Stoner import Data, DataFolder


def test_filedialog():
    def dummy(mode="getOpenFileName"):
        modes = {
            "getOpenFileName": ret_pth,
            "getOpenFileNames": [ret_pth],
            "getSaveFileName": None,
            "getExistingDirectory": ret_pth.parent,
        }
        return lambda *args, **kargs: (modes[mode], None)

    modes = {
        "OpenFile": {
            "method": dummy("getOpenFileName"),
            "caption": "Select file to open...",
            "arg": ["parent", "caption", "directory", "filter", "options"],
        },
        "OpenFiles": {
            "method": dummy("getOpenFileNames"),
            "caption": "Select file(s_ to open...",
            "arg": ["parent", "caption", "directory", "filter", "options"],
        },
        "SaveFile": {
            "method": dummy("getSaveFileName"),
            "caption": "Save file as...",
            "arg": ["parent", "caption", "directory", "filter", "options"],
        },
        "SelectDirectory": {
            "method": dummy("getExistingDirectory"),
            "caption": "Select folder...",
            "arg": ["parent", "caption", "directory", "options"],
        },
    }

    widgets = sys.modules["Stoner.tools.widgets"]
    app = getattr(widgets, "App")
    setattr(app, "modes", modes)

    assert widgets.fileDialog.openDialog() == ret_pth
    assert widgets.fileDialog.openDialog(title="Test", start=".") == ret_pth
    assert widgets.fileDialog.openDialog(patterns={"*.bad": "Very bad files"}) == ret_pth
    assert widgets.fileDialog.openDialog(mode="OpenFiles") == [ret_pth]
    assert widgets.fileDialog.openDialog(mode="SaveFile") is None
    assert widgets.fileDialog.openDialog(mode="SelectDirectory") == ret_pth.parent
    with pytest.raises(ValueError):
        widgets.fileDialog.openDialog(mode="Whateve")


def test_loader():
    d = Data(False)
    assert d.shape == (1676, 3), "Failed to load data with dialog box"
    with pytest.raises(RuntimeError):
        d.save(False)
    fldr = DataFolder(False)
    assert fldr.shape == (
        52,
        {
            "attocube_scan": (15, {}),
            "maximus_scan": (2, {"MPI_210127021": (3, {})}),
            "NLIV": (11, {}),
            "recursivefoldertest": (1, {}),
            "working": (4, {}),
        },
    )
    fldr = DataFolder(False, multifile=True)
    assert fldr.shape == (1, {}), "multifile mode failed!"


def _event(data, name, **kargs):
    """Make a fake event."""
    select = data._select
    event = Event("fake", select.data.fig.canvas)
    for k, v in kargs.items():
        setattr(event, k, v)
    try:
        getattr(select, name)(event)
    except Exception:
        breakpoint()
        pass


def _trigger0(data):
    time.sleep(1)
    select = data._select
    select.onselect(50, 100)
    _event(data, "keypress", key="escape")


def _trigger1(data):
    time.sleep(1)
    select = data._select
    select.onselect(50, 100)
    _event(data, "keypress", key="enter")


def _trigger2(data):
    time.sleep(1)
    select = data._select
    select.onselect(50, 100)
    select.onselect(150, 200)
    _event(data, "keypress", key="i")
    _event(data, "keypress", key="backspace")
    _event(data, "keypress", key="enter")


def test_range_select():
    data = Stoner.Data(ret_pth, setas="xy")
    thread = threading.Thread(target=_trigger0, args=(data,))
    thread.start()
    result = data.search()
    xmin, xmax = result.x.min(), result.x.max()
    assert xmin < 4.4 and xmax > 291, "Failed to select and clear"
    thread = threading.Thread(target=_trigger1, args=(data,))
    thread.start()
    result = data.search()
    xmin1, xmax1 = result.x.min(), result.x.max()
    assert np.isclose(xmin1, 50, atol=1) and np.isclose(xmax1, 100, 1), "Single selection failed."
    thread = threading.Thread(target=_trigger2, args=(data,))
    thread.start()
    result = data.search()
    xmin2, xmax2 = result.x.min(), result.x.max()
    assert np.isclose(xmin, xmin2) and np.isclose(xmax, xmax2), "Selection with keypresses failed"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Test Stoner.tools.tests module."""


import pytest
import numpy as np

from Stoner.tools import tests
from Stoner import Data


def test_all_size():
    x1 = [1] * 5
    x2 = [2] * 5
    assert tests.all_size((x1, x2), 5), "Failed all_size with lists"
    x1 = np.array(x1)
    x2 = np.array(x2)
    assert tests.all_size((x1, x2)), "Failed all_size with arrays"


def test_all_type():
    x1 = [1] * 5
    assert tests.all_type(x1, int), "Failed all_type with list"
    x1 = np.array(x1)
    assert tests.all_type(x1, int), "Failed all type with array"


def test_is_AnyNone():
    x1 = [1] * 5
    x1[3] = None
    assert tests.isanynone(*x1), "is_AbyNione failed."


def test_is_Comparable():
    x1 = np.ones(10)
    x2 = np.zeros(10)
    assert tests.isComparable(x1, x2), "is_Comparable with arrays failed."
    x1 = "Hello"
    x2 = "World"
    assert tests.isComparable(x1, x2), "Failed is_Comparable with type error."
    x1 = 3
    x2 = 5
    assert tests.isComparable(x1, x2), "is_Comparable failed with integers"


def test_is_Iterable():
    assert tests.isiterable(list()), "Is Iterable failed"


def test_is_like_list():
    assert tests.isLikeList(tuple()), "isLikeList failed."


def test_is_None():
    assert tests.isnone(None), "isnone with None failed"
    assert tests.isnone([None, None, None]), "isnone with list oif None failed"

    def g():
        while True:
            yield None

    assert tests.isnone(g()), "isnone with a generator failed"


def test_is_property():
    d = Data()
    assert tests.isproperty(Data, "data"), "Failed to test isProperty with class"
    assert tests.isproperty(d, "data"), "Failed to test isProperty with instance"


def test_is_tuple():
    assert not tests.isTuple(5), "isTuple though 5 was a tuple"
    assert not tests.isTuple((5,), int, int), "isTuple failed to short circuit length check"
    assert tests.isTuple((5,)), "isTuple didn't think (5,) is a tuple"
    assert tests.isTuple((4, "Hi"), int, str), "isTuple failed to match types"
    assert not tests.isTuple((4, "Hi"), str, int), "isTuple failed to match types as bad"


if __name__ == "__main__":
    pytest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implements classes to load file formats from various simulation packages."""

__all__ = ["GenXFile", "OVFFile"]
import re

import numpy as np

from ..Core import DataFile
from ..core.exceptions import StonerLoadError
from ..tools.file import FileManager


def _read_line(data, metadata):
    """Read a single line and add to a metadata dictionary."""
    line = data.readline().decode("utf-8", errors="ignore").strip("#\n \t\r")
    if line == "" or ":" not in line:
        return True
    parts = line.split(":")
    field = parts[0].strip()
    value = ":".join(parts[1:]).strip()
    if field == "Begin" and value.startswith("Data "):
        value = value.split(" ")
        metadata["representation"] = value[1]
        if value[1] == "Binary":
            metadata["representation size"] = value[2]
        return False
    if field not in ["Begin", "End"]:
        metadata[field] = value
    return True


class GenXFile(DataFile):
    """Extends DataFile for GenX Exported data."""

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.dat", "*.txt"]  # Recognised filename patterns

    def _load(self, filename=None, *args, **kargs):
        """Load function. File format has space delimited columns from row 3 onwards."""
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        pattern = re.compile(r'# Dataset "([^\"]*)" exported from GenX3? on (.*)$')
        pattern2 = re.compile(r"#\sFile\sexported\sfrom\sGenX3?\'s\sReflectivity\splugin")
        i = 0
        ix = 0
        with FileManager(self.filename, "r", errors="ignore", encoding="utf-8") as datafile:
            line = datafile.readline()
            match = pattern.match(line)
            match2 = pattern2.match(line)
            if match is not None:
                dataset = match.groups()[0]
                date = match.groups()[1]
                self["date"] = date
                i = 2
            elif match2 is not None:
                line = datafile.readline()
                self["date"] = line.split(":")[1].strip()
                dataset = datafile.readline()[1:].strip()
                i = 3
            else:
                raise StonerLoadError("Not a GenXFile")
            for ix, line in enumerate(datafile):
                line = line.strip()
                if line in ["# Headers:", "# Column lables:"]:
                    line = next(datafile)[1:].strip()
                    break
            else:
                raise StonerLoadError("Cannot find headers")
            skip = ix + i + 2
            column_headers = [f.strip() for f in re.split(r"[\s\t]+", line.strip())]
            self.data = np.real(np.genfromtxt(datafile, dtype=complex))
        self["dataset"] = dataset
        if "sld" in dataset.lower():
            self["type"] = "SLD"
        elif "asymmetry" in dataset.lower():
            self["type"] = "Asymmetry"
        elif "dd" in dataset.lower():
            self["type"] = "Down"
        elif "uu" in dataset.lower():
            self["type"] = "Up"
        self.column_headers = column_headers
        self.setas = "xyye"
        return self


class OVFFile(DataFile):
    """A class that reads OOMMF vector format files and constructs x,y,z,u,v,w data.

    OVF 1 and OVF 2 files with text or binary data and only files with a meshtype rectangular are supported
    """

    #: priority (int): is the load order for the class, smaller numbers are tried before larger numbers.
    #   .. note::
    #      Subclasses with priority<=32 should make some positive identification that they have the right
    #      file type before attempting to read data.
    priority = 16
    #: pattern (list of str): A list of file extensions that might contain this type of file. Used to construct
    # the file load/save dialog boxes.
    patterns = ["*.ovf"]  # Recognised filename patterns

    mime_type = ["text/plain", "application/octet-stream"]

    def _load(self, filename=None, *args, **kargs):
        """Load function. File format has space delimited columns from row 3 onwards.

        Notes:
            This code can handle only the first segment in the data file.
        """
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        # Reading in binary, converting to utf-8 where we should be in the text header
        with FileManager(self.filename, "rb") as data:
            line = data.readline().decode("utf-8", errors="ignore").strip("#\n \t\r")
            if line not in ["OOMMF: rectangular mesh v1.0", "OOMMF OVF 2.0"]:
                raise StonerLoadError("Not an OVF 1.0 or 2.0 file.")
            while _read_line(data, self.metadata):
                pass  # Read the file until we reach the start of the data block.
            if self.metadata["representation"] == "Binary":
                size = (  # Work out the size of the data block to read
                    self.metadata["xnodes"]
                    * self.metadata["ynodes"]
                    * self.metadata["znodes"]
                    * self.metadata["valuedim"]
                    + 1
                ) * self.metadata["representation size"]
                bin_data = data.read(size)
                numbers = np.frombuffer(bin_data, dtype=f"<f{self.metadata['representation size']}")
                chk = numbers[0]
                if (
                    chk != [1234567.0, 123456789012345.0][self.metadata["representation size"] // 4 - 1]
                ):  # If we have a good check number we can carry on, otherwise try the other endianness
                    numbers = np.frombuffer(bin_data, dtype=f">f{self.metadata['representation size']}")
                    chk = numbers[0]
                    if chk != [1234567.0, 123456789012345.0][self.metadata["representation size"] // 4 - 1]:
                        raise StonerLoadError("Bad binary data for ovf gile.")

                data = np.reshape(
                    numbers[1:],
                    (self.metadata["xnodes"] * self.metadata["ynodes"] * self.metadata["znodes"], 3),
                )
            else:
                data = np.genfromtxt(
                    data,
                    max_rows=self.metadata["xnodes"] * self.metadata["ynodes"] * self.metadata["znodes"],
                )
        xmin, xmax, xstep = (
            self.metadata["xmin"],
            self.metadata["xmax"],
            self.metadata["xstepsize"],
        )
        ymin, ymax, ystep = (
            self.metadata["ymin"],
            self.metadata["ymax"],
            self.metadata["ystepsize"],
        )
        zmin, zmax, zstep = (
            self.metadata["zmin"],
            self.metadata["zmax"],
            self.metadata["zstepsize"],
        )
        Z, Y, X = np.meshgrid(
            np.arange(zmin + zstep / 2, zmax, zstep) * 1e9,
            np.arange(ymin + ystep / 2, ymax, ystep) * 1e9,
            np.arange(xmin + xstep / 2, xmax, xstep) * 1e9,
            indexing="ij",
        )
        self.data = np.column_stack((X.ravel(), Y.ravel(), Z.ravel(), data))

        column_headers = ["X (nm)", "Y (nm)", "Z (nm)", "U", "V", "W"]
        self.setas = "xyzuvw"
        self.column_headers = column_headers
        return self



# -*- coding: utf-8 -*-
"""Implements a baseFolder type structure for working with collections of images."""
__all__ = ["ImageFolderMixin", "ImageFolder"]
from warnings import warn
from importlib import import_module
from os import path
from json import loads, dumps
from copy import deepcopy, copy

import numpy as np
from matplotlib.pyplot import figure, Figure, subplot
from PIL.TiffImagePlugin import ImageFileDirectory_v2
from PIL import Image

from .core import ImageArray
from ..Folders import DiskBasedFolderMixin, baseFolder
from ..compat import string_types, int_types
from . import ImageFile


class ImageFolderMixin:
    """Mixin to provide a folder object for images.

    ImageFolderMixin is designed to behave pretty much like DataFolder but with
    functions and loaders appropriate for image based files.

    Attributes:
        type (:py:class:`Stoner.Image.core.ImageArray`):
            the type ob object to store in the folder (defaults to :py:class:`Stoner.Cire.Data`)
        extra_args (dict):
            Extra arguments to use when instantiatoing the contents of the folder from a file on disk.
        pattern (str or regexp):
            A filename globbing pattern that matches the contents of the folder. If a regular expression is
            provided then any named groups are used to construct additional metadata entryies from the filename.
            Default is *.* to match all files with an extension.
        read_means (bool):
            If true, additional metadata keys are added that return the mean value of each column of the data.
            This can hep in grouping files where one column of data contains a constant value for the
            experimental state. Default is False
        recursive (bool):
            Specifies whether to search recursively in a whole directory tree. Default is True.
        flatten (bool):
            Specify where to present subdirectories as separate groups in the folder (False) or as a single group
            (True). Default is False. The :py:meth:`DiskBasedFolderMixin.flatten` method has the equivalent effect and
            :py:meth:`DiskBasedFolderMixin.unflatten` reverses it.
        directory (str):
            The root directory on disc for the folder - by default this is the current working directory.
        multifile (boo):
            Whether to select individual files manually that are not (necessarily) in  a common directory structure.
        readlist (bool):
            Whether to read the directory immediately on creation. Default is True
    """

    _defaults = {"type": ImageArray, "pattern": ["*.png", "*.tiff", "*.jpeg", "*.jpg", "*.tif"]}
    _no_defaults = ["flat"]

    @property
    def size(self):
        """Return the size of an individual image or False if not all images are the same size."""
        if len(self) > 0:
            shape = self[0].shape
        else:
            shape = tuple()
        for i in self:
            if i.shape != shape:
                return False
        return shape

    @property
    def images(self):
        """Iterate over just the images in the Folder."""
        for im in self:
            if not isinstance(im, np.ndarray):
                if hasattr(im, "image"):
                    im = im.image
                else:
                    raise TypeError(f"Cannot represent {type(im)} as an ImageArray.")
            yield im

    #########################################################################################################
    ##### Folder interface methods
    #########################################################################################################

    def __getter__(self, name, instantiate=True):
        """Ensure we set the title on the image.

        Parameters:
           name (key type):
               The canonical mapping key to get the dataObject. By default
               the baseFolder class uses a :py:class:`regexpDict` to store objects in.

        Keyword Arguments:
            instantiate (bool):
                If True (default) then always return a metadataObject. If False, the __getter__ method may return a
                key that can be used by it later to actually get the metadataObject. If None, then will return
                whatever is held in the object cache, either instance
                or name.

        Returns:
            (metadataObject):
                The metadataObject

        Note:
            Mainly we call the parent method and then set the title if it's not already set.'
        """
        ret = super().__getter__(name, instantiate)
        if hasattr(ret, "_title") and ret._title is None:
            ret._title = name
        return ret

    def align(self, *args, **kargs):
        """Align each image in the folder to the reference image.

        Args:
            ref (str, int, ImageFile, ImageArray or 2D array):
                The reference image to align to. If a string or an int, then this is used to lookup the corresponding
                member of the ImageFolder which is then used. ImageFiles, ImageArrays and 2D arrays are used directly
                as reference images.

        Keyword Arguments:
            method (str):
                The method is passed to the :py:class:`Stone.Image.ImageArray.align` method to control how the image
                alignment is done. By default the 'Scharr' method is used.
            box (int, float, tuple of ints or floats):
                Specifies a subset of the images to be used to calculate the alignment with.
            scale (int):
                Magnification factor to scale the image by before doing the alignment for better sub=pixel alignments.

        Returns:
            The aligned ImageFolder.
        """
        if len(args) == 1:
            ref = args[0]
        elif len(args) == 0:
            ref = 0
        else:
            raise ValueError(
                f"{type(self).__name__}.align only takes zero or one positional arguments not {len(args)}!"
            )
        # Get me reference data
        if isinstance(ref, (string_types, int_types)):
            ref_data = self.__getter__(ref, instantiate=True)
            if isinstance(ref_data, ImageFile):
                ref_data = ref_data.image
        elif isinstance(ref, ImageFile):
            ref_data = ref.image.view(ImageArray)
        elif isinstance(ref, np.ndarray) and ref.ndim == 2:
            ref_data = ref.view(ImageArray)
        else:
            try:
                ref_data = np.array(ref).view(ImageArray)
                if ref_data.ndim != 2:
                    raise TypeError()
            except (TypeError, ValueError) as err:
                raise TypeError(f"Cannot interpret {type(ref)} as reference image data.") from err
        # Call align on each object
        self.each.align(ref_data, **kargs)
        limits = self.metadata.slice("translation_limits", output="array")
        stack_limits = np.zeros(4)
        stack_limits[::2] = limits.max(axis=0)[::2]
        stack_limits[1::2] = limits.min(axis=0)[1::2]
        self.metadata["translation_limits"] = tuple(stack_limits)
        stack_limits[::2] = np.ceil(stack_limits)[::2]
        stack_limits[1::2] = np.floor(stack_limits)[1::2]
        self.metadata["align_box"] = tuple(stack_limits.astype(int))
        return self

    def apply_all(self, func, *args, **kargs):
        """Apply function to all images in the stack.

        Args:
            func(string or callable):
                if string it must be a function reachable by ImageArray
            quiet(bool):
                if False print '.' for every iteration

        Note:
            Further args, kargs are passed through to the function
        """
        warn("apply_all is deprecated and will be removed in a future version. Use ImageFolder.each() instead")
        return self.each(func, *args, **kargs)

    def average(self, weights=None, _box=False, _metadata="first"):
        """Get an array of average pixel values for the stack.

        Pass through to numpy average

        Keyword Arguments:
            _box (crop box):
                Specifies the region of the array to be averaged. Default - entire image
            _metadata (str):
                Specifies how to generate metadata for the averaged image.
                - "first": Just ise the first image's metadata
                - "common": Find the common metadata across all images
                - "none': no metadata from images.

        Returns:
            average(ImageArray):
                average values
        """
        if not self.size:
            raise RuntimeError("Cannot average Imagefolder if images have different sizes")
        if hasattr(self, "_stack"):
            stack = self._stack.view(np.ndarray)
            axis = -1
        else:
            stack = np.stack(list(self.images), axis=0)
            axis = 0
        average = np.average(stack, axis=axis, weights=weights)
        ret = average.view(ImageArray)
        if _metadata == "common":
            ret.metadata = self.metadata.common_metadata
        elif _metadata == "first":
            ret.metadata = deepcopy(self[0].metadata)
        return self._type(ret[ret._box(_box)])

    def loadgroup(self):
        """Load all files from this group into memory."""
        for _ in self:
            pass

    def as_stack(self):
        """Return a ImageStack of the images in the current group."""
        stack = import_module(".stack", "Stoner.Image")
        k = stack.ImageStack(self)
        return k

    @classmethod
    def from_tiff(cls, filename, **kargs):
        """Create a new ImageArray from a tiff file."""
        self = cls(**kargs)
        with Image.open(filename, "r") as img:
            tags = img.tag_v2
            if 270 in tags:
                try:
                    userdata = loads(tags[270])
                    typ = userdata.get("type", cls.__name__)
                    mod = userdata.get("module", cls.__module__)
                    layout = userdata.get("layout", (0, {}))

                    mod = import_module(mod)
                    typ = getattr(mod, typ)
                    if not issubclass(typ, ImageFolderMixin):
                        raise TypeError(
                            f"Bad type in Tiff file {typ.__name__} is not a subclass of Stoner.ImageFolder"
                        )
                    metadata = userdata.get("metadata", [])
                except (TypeError, ValueError, IOError):
                    metadata = []
            else:
                raise TypeError("Cannot load as an ImageFolder due to lack of description tag")
            imglist = []
            for ix, md in enumerate(metadata):
                img.seek(ix)
                image = np.asarray(img)
                if image.ndim == 3:
                    if image.shape[2] < 4:  # Need to add a dummy alpha channel
                        image = np.append(np.zeros_like(image[:, :, 0]), axis=2)
                    image = image.view(dtype=np.uint32).reshape(image.shape[:-1])

                if isinstance(self.type, np.ndarray):
                    image = image.view(self.type)
                else:
                    image = self.type(image)
                image.metadata.import_all(md)
                imglist.append(image)

            self._marshall(layout=layout, data=imglist)

        return self

    def mask_select(self):
        """Run the ImageFile.mask.select() on each image."""
        sel = []
        for img in self:
            img.mask.select(_selection=sel)

    def mean(self, _box=False, _metadata="first"):
        """Calculate the mean value of all the images in the stack.

        Keyword Arguments:
            _box (crop box):
                Specifies the region of the array to be averaged. Default - entire image
            _metadata (str):
                Specifies how to generate metadata for the averaged image.
                - "first": Just ise the first image's metadata
                - "common": Find the common metadata across all images
                - "none': no metadata from images.

        Actually a synonym for self.average with not weights
        """
        return self.average(_box=_box, _metadata=_metadata)

    def montage(self, *args, **kargs):
        """Call the plot method for each metadataObject, but switching to a subplot each time.

        Args:
            args: Positional arguments to pass through to the :py:meth:`Stoner.plot.PlotMixin.plot` call.
            kargs: Keyword arguments to pass through to the :py:meth:`Stoner.plot.PlotMixin.plot` call.

        Keyword Arguments:
            extra (callable(i,j,d)):
                A callable that can carry out additional processing per plot after the plot is done
            figsize(tuple(x,y)):
                Size of the figure to create
            dpi(float):
                dots per inch on the figure
            edgecolor,facecolor(matplotlib colour):
                figure edge and frame colours.
            frameon (bool):
                Turns figure frames on or off
            FigureClass(class):
                Passed to matplotlib figure call.
            plots_per_page(int):
                maximum number of plots per figure.

        Returns:
            A list of :py:class:`matplotlib.pyplot.Axes` instances.

        Notes:
            If the underlying type of the :py:class:`Stoner.Core.metadataObject` instances in the
            :py:class:`PlotFolder` lacks a **plot** method, then the instances are converted to
            :py:class:`Stoner.Core.Data`.

            Each plot is generated as sub-plot on a page. The number of rows and columns of subplots is computed
            from the aspect ratio of the figure and the number of files in the :py:class:`PlotFolder`.
        """
        plts = kargs.pop("plots_per_page", getattr(self, "plots_per_page", len(self)))
        plts = min(plts, len(self))

        extra = kargs.pop("extra", lambda i, j, d: None)

        fig_num = kargs.pop("figure", getattr(self, "_figure", None))
        if isinstance(fig_num, Figure):
            kargs.setdefault("figsize", fig_num.get_size_inches())
            kargs.setdefault("facecolor", fig_num.get_facecolor())
            kargs.setdefault("edgecolor", fig_num.get_edgecolor())
            kargs.setdefault("frameon", fig_num.get_frameon())
            kargs.setdefault("FigureClass", fig_num.__class__)
            fig_num = fig_num.number

        fig_args = getattr(self, "_fig_args", [])
        fig_kargs = getattr(self, "_fig_kargs", {"layout": "constrained"})
        for arg in ("figsize", "dpi", "facecolor", "edgecolor", "frameon", "FigureClass"):
            if arg in kargs:
                fig_kargs[arg] = kargs.pop(arg)
        if fig_num is None:
            fig = figure(*fig_args, **fig_kargs)
        else:
            fig = figure(fig_num, **fig_kargs)
        w, h = fig.get_size_inches()
        plt_x = int(np.floor(np.sqrt(plts) * w / h))
        plt_y = int(np.ceil(plts / plt_x))

        kargs["figure"] = fig
        ret = []
        j = 0
        fignum = fig.number
        for i, d in enumerate(self):
            plt_kargs = copy(kargs)
            if i % plts == 0 and i != 0:
                fig = figure(*fig_args, **fig_kargs)
                fignum = fig.number
                j = 1
            else:
                j += 1
            fig = figure(fignum)
            ax = subplot(plt_y, plt_x, j)
            plt_kargs["figure"] = fig
            plt_kargs["ax"] = ax
            if "title" in kargs:
                if isinstance(kargs["title"], str):
                    plt_kargs["title"] = kargs["title"].format(**d)
                elif callable(kargs["title"]):
                    plt_kargs["title"] = kargs["title"](d)
            ret.append(d.imshow(*args, **plt_kargs))
            extra(i, j, d)
        return ret

    def stddev(self, weights=None, _box=False, _metadata="first"):
        """Calculate weighted standard deviation for stack.

        Keyword Arguments:
            _box (crop box):
                Specifies the region of the array to be averaged. Default - entire image
            _metadata (str):
                Specifies how to generate metadata for the averaged image.
                - "first": Just ise the first image's metadata
                - "common": Find the common metadata across all images
                - "none': no metadata from images.

        This is a biased standard deviation, may not be appropriate for small sample sizes
        """
        if weights is None:  # shortcircuit
            if hasattr(self, "_stack"):
                sumsqdev = np.std(self._stack.view(np.ndarray), axis=-1)
            else:
                sumsqdev = np.stack(list(self.images), axis=0).std(axis=0)
        else:
            avs = self.average(weights=weights)
            if not isinstance(avs, np.ndarray) and hasattr(avs, "image"):
                avs = avs.image
            sumsqdev = np.zeros_like(avs)
            for ix, img in enumerate(self.images):
                sumsqdev += weights[ix] * (img - avs) ** 2
            sumsqdev = np.sqrt(sumsqdev) / np.sum(weights, axis=0)
        ret = sumsqdev.view(ImageArray)
        ret.metadata = self.metadata.common_metadata
        return self._type(ret[ret._box(_box)])

    def stderr(self, weights=None, _box=False, _metadata="first"):
        """Calculate standard error in the stack average.

        Keyword Arguments:
            _box (crop box):
                Specifies the region of the array to be averaged. Default - entire image
            _metadata (str):
                Specifies how to generate metadata for the averaged image.
                - "first": Just ise the first image's metadata
                - "common": Find the common metadata across all images
                - "none': no metadata from images.
        """
        serr = self.stddev(weights=weights, _box=_box, _metadata=_metadata) / np.sqrt(len(self))
        return serr

    def to_tiff(self, filename):
        """Save the ImageArray as a tiff image with metadata.

        Args:
            filename (str):
                Filename to save file as.

        Note:
            PIL can save in modes "L" (8bit unsigned int), "I" (32bit signed int),
            or "F" (32bit signed float). In general max info is preserved for "F"
            type so if forcetype is not specified then this is the default. For
            boolean type data mode "L" will suffice and this is chosen in all cases.
            The type name is added as a string to the metadata before saving.

        """
        metadata_export = []
        imlist = []
        for d in self._marshall():
            dtype = np.dtype(d.dtype).name  # string representation of dtype we can save
            d["ImageArray.dtype"] = dtype  # add the dtype to the metadata for saving.
            metadata_export.append(d.metadata.export_all())
            if d.dtype.kind == "b":  # boolean we're not going to lose data by saving as unsigned int
                imlist.append(Image.fromarray(d.image, mode="L"))
            else:
                try:
                    imlist.append(Image.fromarray(d.image))
                except TypeError:
                    imlist.append(Image.fromarray(d.image.astype("float32")))

        ifd = ImageFileDirectory_v2()
        ifd[270] = dumps(
            {
                "type": type(self).__name__,
                "module": type(self).__module__,
                "layout": self.layout,
                "metadata": metadata_export,
            }
        )
        ext = path.splitext(filename)[1]
        if ext in [".tif", ".tiff"]:  # ensure extension is preserved in save
            pass
        else:  # default to tiff
            ext = ".tiff"

        tiffname = path.splitext(filename)[0] + ext
        imlist[0].save(tiffname, save_all=True, append_images=imlist[1:], tiffinfo=ifd)
        return self


class ImageFolder(ImageFolderMixin, DiskBasedFolderMixin, baseFolder):
    """Folder object for images.

    ImageFolder is designed to behave pretty much like DataFolder but with
    functions and loaders appropriate for image based files.

    Attributes:
        type (:py:class:`Stoner.Image.core.ImageArray`):
            the type ob object to store in the folder (defaults to :py:class:`Stoner.Cire.Data`)
        extra_args (dict):
            Extra arguments to use when instantiatoing the contents of the folder from a file on disk.
        pattern (str or regexp):
            A filename globbing pattern that matches the contents of the folder. If a regular expression is provided
            then any named groups are used to construct additional metadata entryies from the filename. Default is *.*
            to match all files with an extension.
        read_means (bool):
            If true, additional metadata keys are added that return the mean value of each column of the data.
            This can hep in grouping files where one column of data contains a constant value for the experimental
            state. Default is False
        recursive (bool):
            Specifies whether to search recursively in a whole directory tree. Default is True.
        flatten (bool):
            Specify where to present subdirectories as separate groups in the folder (False) or as a single group
            (True). Default is False. The :py:meth:`DiskBasedFolderMixin.flatten` method has the equivalent effect
            and :py:meth:`DiskBasedFolderMixin.unflatten` reverses it.
        directory (str):
            The root directory on disc for the folder - by default this is the current working directory.
        multifile (boo):
            Whether to select individual files manually that are not (necessarily) in  a common directory structure.
        readlist (bool):
            Whether to read the directory immediately on creation. Default is True
    """



# -*- coding: utf-8 -*-
"""Specialist functions for dealing with Kerr Images."""

import os
import subprocess  # nosec
import tempfile

import numpy as np
from skimage import exposure, io, transform

from ..core.base import typeHintedDict
from ..compat import which
from ..core.exceptions import assertion, StonerAssertionError

GRAY_RANGE = (0, 65535)  # 2^16
IM_SIZE = (512, 672)  # Standard Kerr image size
AN_IM_SIZE = (554, 672)  # Kerr image with annotation not cropped
pattern_file = os.path.join(os.path.dirname(__file__), "kerr_patterns.txt")

_useful_keys = [
    "X-B-2d",
    "field: units",
    "MicronsPerPixel",
    "Comment:",
    "Contrast Shift",
    "HorizontalFieldOfView",
    "Images to Average",
    "Lens",
    "Magnification",
    "Subtraction Std",
]
_test_keys = ["X-B-2d", "field: units"]  # minimum keys in data to assert that it is a standard file output


def _parse_text(text, key=None):
    """Parse text which has been recognised from an image if key is given specific hints may be applied."""
    # strip any internal white space
    text = [t.strip() for t in text.split()]
    text = "".join(text)

    # replace letters that look like numbers
    errors = [
        ("s", "5"),
        ("S", "5"),
        ("O", "0"),
        ("f", "/"),
        ("::", "x"),
        ("Z", "2"),
        ("l", "1"),
        ("\xe2\x80\x997", "7"),
        ("?", "7"),
        ("I", "1"),
        ("].", "1"),
        ("'", ""),
    ]
    for item in errors:
        text = text.replace(item[0], item[1])

    # apply any key specific corrections
    if key in ["ocr_field", "ocr_scalebar_length_microns"]:
        try:
            text = float(text)
        except ValueError:
            pass  # leave it as string
    # print '{} after processing: \'{}\''.format(key,data)

    return text


def crop_text(kerr_im, copy=False):
    """Crop the bottom text area from a standard Kermit image.

    KeywordArguments:
        copy(bool):
            Whether to return a copy of the data or the original data

    Returns:
    (ImageArray):
        cropped image
    """
    if kerr_im.shape == IM_SIZE:
        return kerr_im
    if kerr_im.shape != AN_IM_SIZE:
        raise ValueError(
            f"Need a full sized Kerr image to crop. Current size is {kerr_im.shape}"
        )  # check it's a normal image
    return kerr_im.crop(None, None, None, IM_SIZE[0], copy=copy)


def reduce_metadata(kerr_im):
    """Reduce the metadata down to a few useful pieces and do a bit of processing.

    Returns:
        (:py:class:`typeHintedDict`): the new metadata
    """
    newmet = {}
    if not all([k in kerr_im.keys() for k in _test_keys]):
        return kerr_im.metadata  # we've not got a standard Labview output, not safe to reduce
    for key in _useful_keys:
        if key in kerr_im.keys():
            newmet[key] = kerr_im[key]
    newmet["field"] = newmet.pop("X-B-2d")  # rename
    if "Subtraction Std" in kerr_im.keys():
        newmet["subtraction"] = newmet.pop("Subtraction Std")
    if "Averaging" in kerr_im.keys():
        if kerr_im["Averaging"]:  # averaging was on
            newmet["Averaging"] = newmet.pop("Images to Average")
        else:
            newmet["Averaging"] = 1
            newmet.pop("Images to Average")
    kerr_im.metadata = typeHintedDict(newmet)
    return kerr_im.metadata


def _tesseract_image(kerr_im, key):
    """Ocr image with tesseract tool.

    im is the cropped image containing just a bit of text
    key is the metadata key we're trying to find, it may give a
    hint for parsing the text generated.
    """
    # first set up temp files to work with
    tmpdir = tempfile.mkdtemp()
    textfile = os.path.join(tmpdir, "tmpfile.txt")
    stdoutfile = os.path.join(tmpdir, "logfile.txt")
    imagefile = os.path.join(tmpdir, "tmpim.tif")
    with open(textfile, "w", encoding="utf-8") as tf:  # open a text file to export metadata to temporarily
        pass

    # process image to make it easier to read
    i = 1.0 * kerr_im / np.max(kerr_im)  # change to float and normalise
    i = exposure.rescale_intensity(i, in_range=(0.49, 0.5))  # saturate black and white pixels
    i = exposure.rescale_intensity(i)  # make sure they're black and white
    i = transform.rescale(i, 5.0, mode="constant")  # rescale to get more pixels on text
    io.imsave(
        imagefile, (255.0 * i).astype("uint8"), plugin="pil"
    )  # python imaging library will save according to file extension

    # call tesseract
    if kerr_im.tesseractable:
        tesseract = which("tesseract")
        with open(stdoutfile, "w", encoding="utf-8") as stdout:
            subprocess.call(  # nosec
                [tesseract, imagefile, textfile[:-4]], stdout=stdout, stderr=subprocess.STDOUT
            )  # adds '.txt' extension itkerr_im
        os.unlink(stdoutfile)
    with open(textfile, "r", encoding="utf-8") as tf:
        data = tf.readline()

    # delete the temp files
    os.remove(textfile)
    os.remove(imagefile)
    os.rmdir(tmpdir)

    # parse the reading
    if len(data) == 0:
        print(f"No data read for {key}")
    data = _parse_text(data, key=key)
    return data


def get_scalebar(kerr_im):
    """Get the length in pixels of the image scale bar."""
    im = kerr_im[519:520, :419]
    im = im.astype(float)
    im = (im - im.min()) / (im.max() - im.min())
    im = exposure.rescale_intensity(im, in_range=(0.49, 0.5))  # saturate black and white pixels
    im = exposure.rescale_intensity(im)  # make sure they're black and white
    im = np.diff(im[0])  # 1d numpy array, differences
    lim = [np.where(im > 0.9)[0][0], np.where(im < -0.9)[0][0]]  # first occurrence of both cases
    assertion(len(lim) == 2, "Couldn't find scalebar")
    return lim[1] - lim[0]


def float_and_croptext(kerr_im):
    """Convert image to float and crop_text.

    Just to group typical functions together
    """
    ret = kerr_im.asfloat()
    ret = ret.crop_text()
    return ret


def ocr_metadata(kerr_im, field_only=False):
    """Use image recognition to try to pull the metadata numbers off the image.

    Requirements:
        This function uses tesseract to recognise the image, therefore
        tesseract file1 file2 must be valid on your command line.
        Install tesseract from
        https://sourceforge.net/projects/tesseract-ocr-alt/files/?source=navbar

    KeywordArguments:
        field_only(bool):
            only try to return a field value

    Returns:
        metadata: dict
            updated metadata dictionary
    """
    if kerr_im.shape != AN_IM_SIZE:
        pass  # can't do anything without an annotated image

    # now we have to crop the image to the various text areas and try tesseract
    elif field_only:
        fbox = (110, 165, 527, 540)  # (This is just the number area not the unit)
        im = kerr_im.crop(box=fbox, copy=True)
        field = kerr_im._tesseract_image(im, "ocr_field")
        kerr_im.metadata["ocr_field"] = field
    else:
        text_areas = {
            "ocr_field": (110, 165, 527, 540),
            "ocr_date": (542, 605, 512, 527),
            "ocr_time": (605, 668, 512, 527),
            "ocr_subtract": (237, 260, 527, 540),
            "ocr_average": (303, 350, 527, 540),
        }
        try:
            sb_length = kerr_im.get_scalebar()
        except (StonerAssertionError, AssertionError):
            sb_length = None
        if sb_length is not None:
            text_areas.update(
                {
                    "ocr_scalebar_length_microns": (sb_length + 10, sb_length + 27, 514, 527),
                    "ocr_lens": (sb_length + 51, sb_length + 97, 514, 527),
                    "ocr_zoom": (sb_length + 107, sb_length + 149, 514, 527),
                }
            )

        metadata = {}  # now go through and process all keys
        for key in text_areas:
            im = kerr_im.crop(box=text_areas[key], copy=True)
            metadata[key] = _tesseract_image(kerr_im, key)
        metadata["ocr_scalebar_length_pixels"] = sb_length
        if isinstance(metadata["ocr_scalebar_length_microns"], float):
            metadata["ocr_microns_per_pixel"] = metadata["ocr_scalebar_length_microns"] / sb_length
            metadata["ocr_pixels_per_micron"] = 1 / metadata["ocr_microns_per_pixel"]
            metadata["ocr_field_of_view_microns"] = np.array(IM_SIZE) * metadata["ocr_microns_per_pixel"]
        kerr_im.metadata.update(metadata)
    if "ocr_field" in kerr_im.metadata.keys() and not isinstance(kerr_im.metadata["ocr_field"], (int, float)):
        kerr_im.metadata["ocr_field"] = np.nan  # didn't read the field properly
    return kerr_im.metadata


def defect_mask(kerr_im, thresh=0.6, corner_thresh=0.05, radius=1, return_extra=False):
    """Try to create a boolean array which is a mask for typical defects found in Image images.

    Best for unprocessed raw images. (for subtract images
    see defect_mask_subtract_image)
    Looks for big bright things by thresholding and small and dark defects using
    skimage's corner_fast algorithm

    Parameters:
    thresh (float):
        brighter stuff than this gets removed (after image levelling)
    corner_thresh (float):
        see corner_fast (skimage):
    radius (float):
        radius of pixels around corners that are added to mask
    return_extra (bool):
        this returns a dictionary with some of the intermediate steps of the
        calculation

    Returns:
        totmask (ndarray of bool):
            mask
    info (*optional* dict):
        dictionary of intermediate calculation steps
    """
    im = kerr_im.asfloat()
    im = im.level_image(poly_vert=3, poly_horiz=3)
    th = im.threshold_minmax(0, thresh)
    # corner fast good at finding the small black dots
    cor = im.corner_fast(threshold=corner_thresh)
    blobs = cor.blob_doh(min_sigma=1, max_sigma=20, num_sigma=3, threshold=0.01)
    q = np.zeros_like(im)
    for y, x, _ in blobs:
        q[
            int(np.round(y - radius)) : int(np.round(y + radius)),
            int(np.round(x - radius)) : int(np.round(x + radius)),
        ] = 1.0
    totmask = np.logical_or(q, th)
    if return_extra:
        info = {"flattened_image": im, "corner_fast": cor, "corner_points": blobs, "corner_mask": q, "thresh_mask": th}
        return totmask, info
    return totmask


def defect_mask_subtract_image(kerr_im, threshmin=0.25, threshmax=0.9, denoise_weight=0.1, return_extra=False):
    """Create a mask array for a typical subtract Image image.

    Uses a denoise algorithm followed by simple thresholding.

    Returns:
        totmask (ndarray of bool):
            the created mask
        info (*optional* dict):
            the intermediate denoised image
    """
    p = kerr_im.denoise_tv_chambolle(weight=denoise_weight)
    submask = p.threshold_minmax(threshmin, threshmax)
    if return_extra:
        info = {"denoised_image": p}
        return submask, info
    return submask



# -*- coding: utf-8 -*-
"""
test_Folders.py

Created on Mon Jul 18 14:13:39 2016

@author: phygbu
"""


import sys
import os.path as path
import os
import numpy as np
import re
import fnmatch
from numpy import ceil
from copy import copy
from Stoner.compat import Hyperspy_ok
import pytest
from Stoner import DataFolder, __homepath__
from Stoner.folders import PlotFolder

from Stoner import Data
from Stoner.core.base import regexpDict
from Stoner.folders.core import baseFolder
import matplotlib.pyplot as plt

import tempfile

pth = __homepath__ / ".."
sys.path.insert(0, pth)

"""Path to sample Data File"""
datadir = pth / "sample-data"


def test_Folders():
    fldr = DataFolder(datadir, debug=False, recursive=False)
    if not Hyperspy_ok:
        del fldr[".*emd$"]
    fl = len(fldr)
    skip = 1 if Hyperspy_ok else 2
    datfiles = fnmatch.filter(os.listdir(datadir), "*.dat")
    length = (
        len([i for i in os.listdir(datadir) if path.isfile(os.path.join(datadir, i))]) - skip
    )  # don't coiunt TDMS index
    assert length == fl, "Failed to initialise DataFolder from sample data {} {} {} {}".format(
        fl, length, skip, Hyperspy_ok
    )
    assert fldr.index(path.basename(fldr[-1].filename)) == fl - 1, "Failed to index back on filename"
    assert fldr.count(path.basename(fldr[-1].filename)) == 1, "Failed to count filename with string"
    assert fldr.count("*.dat") == len(datfiles), "Count with a glob pattern failed"
    assert len(fldr[::2]) == ceil(len(fldr) / 2.0), "Failed to get the correct number of elements in a folder slice"


def test_loader_opts():
    fldr7 = DataFolder(
        path.join(datadir, "NLIV"), pattern=re.compile(r".*at (?P<field>[0-9\-\.]*)\.txt"), read_means=True
    )
    x = fldr7.metadata.slice(["field", "Voltage", "Current"], output="Data")
    assert x.span("field") == (-0.05, 0.04), "Extract from name pattern and slice into metadata failed."
    assert all(x // "Current" < 0) and all(x // "Current" > -1e-20), "Extract means failed."
    assert list(fldr7.not_loaded) == [], "Not loaded attribute failed."
    fldr7.unload(0)
    assert len(list(fldr7.not_loaded)) == 1, "Unload by index failed."
    fldr7.unload()
    assert len(list(fldr7.not_loaded)) == len(fldr7), "Unload all failed."

    def add_col(d):
        d.add_column(np.ones(len(d)) * d["field"], header="field")

    fldr7.each(add_col)
    fldr7.concatenate()
    assert fldr7[0].shape == (909, 4), "Concatenate failed."


def test_groups_methods():
    fldr = DataFolder(datadir, debug=False, recursive=False)
    if not Hyperspy_ok:
        del fldr[".*emd$"]
    fldr.group("Loaded as")
    fldr.groups.keep(["QDFile", "OpenGDAFile"])
    assert fldr.shape == (0, {"OpenGDAFile": (1, {}), "QDFile": (4, {})}), "groups.keep method failed on folder"


def test_discard_earlier():
    fldr2 = DataFolder(path.join(pth, "tests/Stoner/folder_data"), pattern="*.dat", discard_earlier=True)
    fldr3 = DataFolder(path.join(pth, "tests/Stoner/folder_data"), pattern="*.dat")
    assert len(fldr2) == 1, "Folder created with disacrd_earlier has wrong length ({})".format(len(fldr2))
    assert len(fldr3) == 5, "Folder created without disacrd_earlier has wrong length ({})".format(len(fldr3))
    fldr3.keep_latest()
    assert list(fldr2.ls) == list(
        fldr3.ls
    ), "Folder.keep_latest didn't do the same as discard_earliest in constructor."


def test_clear_and_attrs():
    fldr = DataFolder(datadir, debug=False, recursive=False)
    if not Hyperspy_ok:
        del fldr[".*emd$"]
    fldr2 = fldr.clone
    fldr2.clear()
    assert fldr2.shape == (0, {}), "Failed to clear"
    fldr2.files = fldr.files
    fldr2.groups = fldr.groups
    assert fldr2.shape == fldr.shape, "Failed to write to files and groups attri"
    fldr.each.debug = True
    assert fldr[0].debug, "Setting an attribute on fldr didn't propagate to the contents"
    del fldr.each.debug
    assert not hasattr(fldr[0], "hello"), "Failed to delete attribute from DataFolder"
    with pytest.raises(AttributeError):
        del fldr.debug


def test_Operators():
    fldr = DataFolder(datadir, debug=False, recursive=False)
    if not Hyperspy_ok:
        del fldr[".*emd$"]
    fl = len(fldr)
    d = Data(np.ones((100, 5)))
    fldr += d
    assert fl + 1 == len(fldr), "Failed += operator on DataFolder"
    fldr2 = fldr + fldr
    assert (fl + 1) * 2 == len(fldr2), "Failed + operator with DataFolder on DataFolder"
    fldr -= "Untitled"
    assert len(fldr) == fl, "Failed to remove Untitled-0 from DataFolder by name."
    fldr -= "New-XRay-Data.dql"
    assert fl - 1 == len(fldr), "Failed to remove NEw Xray data by name."
    fldr += "New-XRay-Data.dql"
    assert len(fldr) == fl, "Failed += operator with string on DataFolder"
    fldr /= "Loaded as"
    assert len(fldr["QDFile"]) == 4, "Failoed to group folder by Loaded As metadata with /= operator."
    assert isinstance(fldr["QDFile", "Byapp"], Data), "Indexing group and then common metadata failed"
    fldr = DataFolder(datadir, debug=False, recursive=False)
    fldr2 = DataFolder(path.join(datadir, "NLIV"), pattern="*.txt")
    fldr2.group(lambda x: "zero" if x["iterator"] % 2 == 0 else "one")
    fldr3 = fldr + fldr2
    assert fldr3.shape == (52, {"one": (9, {}), "zero": (7, {})}), "Adding two DataFolders with groups failed"
    fldr4 = fldr3 - fldr2
    fldr4.prune()
    assert fldr4.shape == fldr.shape, "Failed to subtract one DataFolder from another :{}".format(fldr4.shape)
    del fldr2["one"]
    assert fldr2.shape == (0, {"zero": (7, {})}), "Delitem with group failed"
    fldr2.key = path.basename(fldr2.key)
    assert repr(fldr2) == (
        "DataFolder(NLIV) with pattern ('*.txt',) has 0 files and 1 groups\n\tDataFolder(zero) with pattern "
        + "['*.txt'] has 7 files and 0 groups"
    ), "Representation methods failed"
    fldr = DataFolder(datadir, debug=False, recursive=False)
    names = list(fldr.ls)[::2]
    fldr -= names
    assert len(fldr) == 26, "Failed to delete from a sequence"
    with pytest.raises(TypeError):
        fldr - 0.34
    with pytest.raises(RuntimeError):
        fldr - Data()
    with pytest.raises(RuntimeError):
        fldr - "Wiggle"


def test_Base_Operators():
    fldr = DataFolder(datadir, debug=False, recursive=False)
    if not Hyperspy_ok:
        del fldr[".*emd$"]
    for d in fldr:
        _ = d["Loaded as"]
    fldr = baseFolder(fldr)
    fl = len(fldr)
    d = Data(np.ones((100, 5)))
    fldr += d
    assert fl + 1 == len(fldr), "Failed += operator on DataFolder"
    fldr2 = fldr + fldr
    assert (fl + 1) * 2 == len(fldr2), "Failed + operator with DataFolder on DataFolder"
    fldr -= "Untitled"
    assert len(fldr) == fl, "Failed to remove Untitled-0 from DataFolder by name."
    fldr -= "New-XRay-Data.dql"
    assert fl - 1 == len(fldr), "Failed to remove NEw Xray data by name."
    if Hyperspy_ok:
        del fldr["1449 37.0 kx.emd"]
    fldr /= "Loaded as"
    assert len(fldr["QDFile"]) == 4, "Failoed to group folder by Loaded As metadata with /= operator."
    fldr = DataFolder(datadir, debug=False, recursive=False)
    for d in fldr:
        _ = d["Loaded as"]
    fldr = baseFolder(fldr)
    fldr2 = DataFolder(path.join(datadir, "NLIV"), pattern="*.txt")
    fldr2.group(lambda x: "zero" if x["iterator"] % 2 == 0 else "one")
    fldr3 = fldr + fldr2
    assert fldr3.shape == (52, {"one": (9, {}), "zero": (7, {})}), "Adding two DataFolders with groups failed"
    fldr4 = fldr3 - fldr2
    fldr4.prune()
    assert fldr4.shape == fldr.shape, "Failed to subtract one DataFolder from another :{}".format(fldr4.shape)
    del fldr2["one"]
    assert fldr2.shape == (0, {"zero": (7, {})}), "Delitem with group failed"
    fldr2.key = path.basename(fldr2.key)
    assert repr(fldr2) == (
        "DataFolder(NLIV) with pattern ('*.txt',) has 0 files and 1 groups\n\tDataFolder(zero) with pattern"
        + " ['*.txt'] has 7 files and 0 groups"
    ), "Representation methods failed"
    fldr = DataFolder(datadir, debug=False, recursive=False)
    names = list(fldr.ls)[::2]
    fldr -= names
    assert len(fldr) == 26, "Failed to delete from a sequence"
    with pytest.raises(TypeError):
        fldr - 0.34
    with pytest.raises(RuntimeError):
        fldr - Data()
    with pytest.raises(RuntimeError):
        fldr - "Wiggle"


def test_Properties():
    fldr = DataFolder(datadir, debug=False, recursive=False)
    if not Hyperspy_ok:
        del fldr[".*emd$"]
    assert fldr.mindepth == 0, "Minimum depth of flat group n ot equal to zero."
    fldr /= "Loaded as"
    grps = list(fldr.lsgrp)
    skip = 0 if Hyperspy_ok else 1
    assert len(grps) == 26 - skip, f"Length of lsgrp not as expected: {len(grps)} not {27-skip}"
    fldr.debug = True
    fldr = fldr
    assert fldr["XRDFile"][0].debug, "Setting debug on folder failed!"
    fldr.debug = False
    fldr["QDFile"].group("Byapp")
    assert fldr.trunkdepth == 1, "Trunkdepth failed"
    assert fldr.mindepth == 1, "mindepth attribute of folder failed."
    assert fldr.depth == 2, "depth attribute failed."
    fldr = DataFolder(datadir, debug=False, recursive=False)
    fldr += Data()
    skip = 1 if Hyperspy_ok else 2
    assert len(list(fldr.loaded)) == 1, "loaded attribute failed {}".format(len(list(fldr.loaded)))
    assert len(list(fldr.not_empty)) == len(fldr) - skip, "not_empty attribute failed."
    fldr -= "Untitled"
    assert not fldr.is_empty, "fldr.is_empty failed"
    fldr = DataFolder(datadir, debug=False, recursive=False)
    objects = copy(fldr.objects)
    fldr.objects = dict(objects)
    assert isinstance(fldr.objects, regexpDict), "Folder objects not reset to regexp dictionary"
    fldr.objects = objects
    assert isinstance(fldr.objects, regexpDict), "Setting Folder objects mangled type"
    fldr.type = Data()
    assert issubclass(fldr.type, Data), "Setting type by instance of class failed"


def test_methods():
    sliced = np.array(
        [
            "DataFile",
            "MDAASCIIFile",
            "BNLFile",
            "DataFile",
            "DataFile",
            "DataFile",
            "DataFile",
            "DataFile",
            "MokeFile",
            "EasyPlotFile",
            "DataFile",
            "DataFile",
            "DataFile",
        ],
        dtype="<U12",
    )
    fldr = DataFolder(datadir, pattern="*.txt", recursive=False).sort()

    test_sliced = fldr.slice_metadata("Loaded as")
    assert len(sliced) == len(test_sliced), "Test slice not equal length - sample-data changed? {}".format(
        test_sliced
    )
    assert np.all(test_sliced == sliced), "Slicing metadata failed to work."
    fldr.insert(5, Data())
    assert list(fldr.ls)[5] == "Untitled", "Insert failed"
    fldr = fldr
    _ = fldr[-1]
    assert list(reversed(fldr))[0].filename == fldr[-1].filename


def test_clone():
    fldr = DataFolder(datadir, pattern="*.txt")
    fldr.abc = 123  # add an attribute
    t = fldr.__clone__()
    assert t.pattern == fldr.pattern, "pattern didn't copy over"
    assert hasattr(t, "abc") and t.abc == 123, "user attribute didn't copy over"
    assert isinstance(t["recursivefoldertest"], DataFolder), "groups didn't copy over"


def test_grouping():
    fldr4 = DataFolder()
    x = np.linspace(-np.pi, np.pi, 181)
    for phase in np.linspace(0, 1.0, 5):
        for amplitude in np.linspace(1, 2, 6):
            for frequency in np.linspace(1, 2, 5):
                y = amplitude * np.sin(frequency * x + phase * np.pi)
                d = Data(x, y, setas="xy", column_headers=["X", "Y"])
                d["frequency"] = frequency
                d["amplitude"] = amplitude
                d["phase"] = phase
                d["params"] = [phase, frequency, amplitude]
                d.filename = "test/{amplitude}/{phase}/{frequency}.dat".format(**d)
                fldr4 += d
    fldr4.unflatten()
    assert fldr4.mindepth == 3, "Unflattened DataFolder had wrong mindepth."
    assert fldr4.shape == (~~fldr4).shape, "Datafodler changed shape on flatten/unflatten"
    fldr5 = fldr4.select(amplitude=1.4, recurse=True)
    fldr5.prune()
    pruned = (
        0,
        {
            "test": (
                0,
                {"1.4": (0, {"0.0": (5, {}), "0.25": (5, {}), "0.5": (5, {}), "0.75": (5, {}), "1.0": (5, {})})},
            )
        },
    )
    selected = (0, {"test": (0, {"1.4": (0, {"0.25": (1, {}), "0.5": (1, {}), "0.75": (1, {}), "1.0": (1, {})})})})
    assert fldr5.shape == pruned, "Folder pruning gave an unexpected shape."
    assert fldr5[("test", "1.4", "0.5", 0, "phase")] == 0.5, "Multilevel indexing of tree failed."
    shape = (~(~fldr4).select(amplitude=1.4).select(frequency=1).select(phase__gt=0.2)).shape
    fldr4 = fldr4
    assert shape == selected, "Multi selects and inverts failed."
    g = (~fldr4) / 10
    assert g.shape == (
        0,
        {
            "Group 0": (15, {}),
            "Group 1": (15, {}),
            "Group 2": (15, {}),
            "Group 3": (15, {}),
            "Group 4": (15, {}),
            "Group 5": (15, {}),
            "Group 6": (15, {}),
            "Group 7": (15, {}),
            "Group 8": (15, {}),
            "Group 9": (15, {}),
        },
    ), "Dive by int failed."
    g["Group 6"] -= 5
    assert g.shape == (
        0,
        {
            "Group 0": (15, {}),
            "Group 1": (15, {}),
            "Group 2": (15, {}),
            "Group 3": (15, {}),
            "Group 4": (15, {}),
            "Group 5": (15, {}),
            "Group 6": (14, {}),
            "Group 7": (15, {}),
            "Group 8": (15, {}),
            "Group 9": (15, {}),
        },
    ), "Sub by int failed."
    remove = g["Group 3"][4]
    g["Group 3"] -= remove
    assert g.shape == (
        0,
        {
            "Group 0": (15, {}),
            "Group 1": (15, {}),
            "Group 2": (15, {}),
            "Group 3": (14, {}),
            "Group 4": (15, {}),
            "Group 5": (15, {}),
            "Group 6": (14, {}),
            "Group 7": (15, {}),
            "Group 8": (15, {}),
            "Group 9": (15, {}),
        },
    ), "Sub by object failed."
    d = fldr4["test", 1.0, 1.0].gather(0, 1)
    assert d.shape == (181, 6), "Gather seems have failed."
    assert np.all(fldr4["test", 1.0, 1.0].slice_metadata("phase") == np.ones(5)), "Slice metadata failure."
    d = (~fldr4).extract("phase", "frequency", "amplitude", "params")
    assert d.shape == (150, 6), "Extract failed to produce data of correct shape."
    assert d.column_headers == [
        "phase",
        "frequency",
        "amplitude",
        "params",
        "params",
        "params",
    ], "Extract failed to get correct column headers."
    p = fldr4["test", 1.0, 1.0]
    p = PlotFolder(p)
    p.plot()
    assert len(plt.get_fignums()) == 1, "Failed to generate a single plot for PlotFolder."
    plt.close("all")


def test_saving():
    fldr4 = DataFolder()
    x = np.linspace(-np.pi, np.pi, 181)
    for phase in np.linspace(0, 1.0, 5):
        for amplitude in np.linspace(1, 2, 6):
            for frequency in np.linspace(1, 2, 5):
                y = amplitude * np.sin(frequency * x + phase * np.pi)
                d = Data(x, y, setas="xy", column_headers=["X", "Y"])
                d["frequency"] = frequency
                d["amplitude"] = amplitude
                d["phase"] = phase
                d["params"] = [phase, frequency, amplitude]
                d.filename = "test/{amplitude}/{phase}/{frequency}.dat".format(**d)
                fldr4 += d
    fldr4.unflatten()
    newdir = tempfile.mkdtemp()
    fldr4.save(newdir)
    fldr5 = DataFolder(newdir)
    assert fldr4.shape == fldr5.shape, "Saved DataFolder and loaded DataFolder have different shapes"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



"""Example of Arrhenius Fit."""

# pylint: disable=invalid-name
from numpy import linspace, ceil, log10, abs as np_abs
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.thermal import Arrhenius, arrhenius

# Make some data
T = linspace(200, 350, 101)
R = arrhenius(T + normal(size=len(T), scale=3.0, loc=0.0), 1e6, 0.5)
E = 10 ** ceil(log10(np_abs(R - arrhenius(T, 1e6, 0.5))))
d = Data(T, R, E, setas="xye", column_headers=["T", "Rate"])

# Curve_fit on its own
d.curve_fit(arrhenius, p0=(1e6, 0.5), result=True, header="curve_fit")
d.setas = "xyey"
d.plot(fmt=["r.", "b-"])
d.annotate_fit(
    arrhenius,
    x=0.5,
    y=0.5,
    mode="eng",
    fontdict={"size": "x-small", "color": "blue"},
)

# lmfit using lmfit guesses
fit = Arrhenius()
d.setas = "xye"
d.lmfit(fit, result=True, header="lmfit")
d.setas = "x...y"
d.plot(fmt="g-")
d.annotate_fit(
    Arrhenius,
    x=0.5,
    y=0.35,
    prefix="Arrhenius",
    mode="eng",
    fontdict={"size": "x-small", "color": "green"},
)

d.setas = "xye"
res = d.odr(Arrhenius, result=True, header="odr", prefix="ODR")
d.setas = "x....y"
d.plot(fmt="m-")
d.annotate_fit(
    Arrhenius,
    x=0.5,
    y=0.2,
    prefix="ODR",
    mode="eng",
    fontdict={"size": "x-small", "color": "magenta"},
)


d.title = "Arrhenius Test Fit"
d.ylabel = "Rate"
d.xlabel = "Temperature (K)"
d.yscale = "log"



# -*- coding: utf-8 -*-
"""Demonstrate PlotFolder Class."""
# pylint: disable=invalid-name
from os.path import join

from numpy import where
from scipy.constants import mu_0
from scipy.stats import gmean

from Stoner import Data, __home__
from Stoner.analysis.fitting.models.magnetism import FMR_Power, Inverse_Kittel
from Stoner.analysis.fitting.models.generic import Linear
from Stoner.plot.formats import DefaultPlotStyle, TexEngFormatter
from Stoner.Folders import PlotFolder

# Customise a plot template
template = DefaultPlotStyle()
template.xformatter = TexEngFormatter()
template.xformatter.unit = "T"
template.yformatter = TexEngFormatter


def field_sign(r):
    """Return string key for sign of field.."""
    pos = r["Field"] >= 0
    return where(pos, "pos", "neg")


def extra(_, __, d):
    """Customise each individual plot."""
    d.axvline(x=d["cut"], ls="--")
    d.title = r"$\nu={:.1f}\,$GHz".format(d.mean("Frequency") / 1e9)
    d.xlabel = r"Field $\mu_0H\,$"
    d.ylabel = "Abs. (arb)"
    d.plt_legend(loc=3)
    d.annotate_fit(FMR_Power, fontdict={"size": 8}, x=0.05, y=0.25)


def do_fit(f):
    """Fit just one set of data."""
    f.template = template
    f["cut"] = f.threshold(1.75e5, rising=False, falling=True)
    f["Frequency"] = (f // "Frequency").mean()
    f.lmfit(
        FMR_Power, result=True, header="Fit", bounds=lambda x, r: x < f["cut"]
    )
    return f


###############################################################################################
#################### Important - if using multiprocessing on Windows, this  block must be ######
#################### Inside a if __name__=="__main__": test. ###################################
###############################################################################################

if __name__ == "__main__":
    # Load data
    d = Data(join(__home__, "..", "sample-data", "FMR-data.txt"))
    # Rename columns and reset plot labels
    d.rename("multi[1]:y", "Field").rename("multi[0]:y", "Frequency").rename(
        "Absorption::X", "FMR"
    )
    d.labels = None

    # Define x and y columns and normalise to a big number
    d.setas(x="Field", y="FMR")  # pylint: disable=not-callable
    d.normalise(base=(-1e6, 1e6))
    fldr = d.split(field_sign, "Frequency")

    # Split the data file into separate files by frequencies and sign of field
    fldr = PlotFolder(fldr)  # Convert to a PlotFolder
    fldr.template = template  # Set my custom plot template
    for f in fldr["neg"]:  # Invert the negative field side
        f.x = -f.x[::-1]
        f.y = -f.y[::-1]

    resfldr = (
        PlotFolder()
    )  # Somewhere to keep the results from +ve and -ve fields

    for s in fldr.groups:  # Fit each FMR spectra
        subfldr = fldr[s]
        subfldr.metadata["Field Sign"] = s

        print("s={}".format(s))
        (do_fit @ subfldr.each)()
        result = subfldr.metadata.slice("Frequency", FMR_Power, output="Data")
        result.metadata.update(
            {x: subfldr[0][x] for x in subfldr.metadata.common_keys}
        )

        # Now plot all the fits
        subfldr.setas[12] = "y"
        subfldr.plots_per_page = 6  # Plot results
        subfldr.plot(
            figsize=(8, 8), fmt=["k.", "r-"], markersize=1, extra=extra
        )

        # Work with the overall results
        try:
            result.setas(  # pylint: disable=not-callable
                y="H_res$", e="H_res err", x="Frequency"
            )  # pylint: disable=not-callable
            result.y = result.y / mu_0  # Convert to A/m
            result.e = result.e / mu_0

            resfldr += result  # Stash the results
        except KeyError:
            continue

    # Merge the two field signs into a single file, taking care of the error columns too
    result = resfldr[0].clone
    for c in [0, 2, 4, 6, 8]:
        result.data[:, c] = (resfldr[1][:, c] + resfldr[0][:, c]) / 2.0
    for c in [1, 3, 5, 7]:
        result.data[:, c] = gmean((resfldr[0][:, c], resfldr[1][:, c]), axis=0)

    # Doing the Kittel fit with an orthogonal distance regression as we have x errors not y errors
    p0 = [2, 200e3, 10e3]  # Some sensible guesses
    result.lmfit(
        Inverse_Kittel, p0=p0, result=True, header="Kittel Fit", output="report"
    )
    result.setas[-1] = "y"

    result.template.yformatter = TexEngFormatter
    result.template.xformatter = TexEngFormatter
    result.labels = None
    result.figure(figsize=(6, 8), no_axes=True)
    result.subplot(211)
    result.plot(fmt=["r.", "b-"])
    result.annotate_fit(Inverse_Kittel, x=7e9, y=1e5, fontdict={"size": 8})
    result.ylabel = "$H_{res} \\mathrm{(Am^{-1})}$"
    result.title = "Inverse Kittel Fit"

    # Get alpha
    result.subplot(212)
    result.setas(y="Delta_H", e="Delta_H err", x="Freq")
    result.y /= mu_0
    result.e /= mu_0
    result.lmfit(Linear, result=True, header="Width", output="report")
    result.setas[-1] = "y"
    result.plot(fmt=["r.", "b-"])
    result.annotate_fit(Linear, x=5.5e9, y=2.8e3, fontdict={"size": 8})



"""Stoner .Analysis provides a subclass of :class:`.Data` that has extra analysis routines builtin."""

__all__ = ["AnalysisMixin", "GetAffineTransform", "ApplyAffineTransform"]
from inspect import getfullargspec
import numpy as np
import numpy.ma as ma

from scipy.integrate import cumtrapz
from scipy.optimize import curve_fit

from .tools import isiterable, isTuple
from .compat import string_types
from .core.exceptions import assertion
from .analysis.utils import threshold as _threshold, ApplyAffineTransform, GetAffineTransform

# from matplotlib.pylab import * #Surely not?


class AnalysisMixin:
    """A mixin calss designed to work with :py:class:`Stoner.Core.DataFile` to provide additional analysis methods."""

    def __dir__(self):
        """Handle the local attributes as well as the inherited ones."""
        attr = dir(type(self))
        attr.extend(super().__dir__())
        attr.extend(list(self.__dict__.keys()))
        attr = list(set(attr))
        return sorted(attr)

    def apply(self, func, col=None, replace=True, header=None, **kargs):
        """Apply the given function to each row in the data set and adds to the data set.

        Args:
            func (callable):
                The function to apply to each row of the data.
            col (index):
                The column in which to place the result of the function

        Keyword Arguments:
            replace (bool):
                Either replace the existing column/complete data or create a new column or data file.
            header (string or None):
                The new column header(s) (defaults to the name of the function func

        Note:
            If any extra keyword arguments are supplied then these are passed to the function directly. If
            you need to pass any arguments that overlap with the keyword arguments to :py:math:`AnalysisMixin.apply`
            then these can be supplied in a dictionary argument *_extra*.

            The callable *func* should have a signature::

                def func(row,**kargs):

            and should return either a single float, in which case it will be used to repalce the specified column,
            or an array, in which case it is used to completely replace the row of data.

            If the function returns a complete row of data, then the *replace* parameter will cause the return
            value to be a new datafile, leaving the original unchanged. The *headers* parameter can give the complete
            column headers for the new data file.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.
        """
        if col is None:
            col = self.setas.get("y", [0])[0]
        col = self.find_col(col)
        kargs.update(kargs.pop("_extra", dict()))
        # Check the dimension of the output
        ret = func(next(self.rows()), **kargs)
        try:
            next(self.rows(reset=True))
        except (RuntimeError, StopIteration):
            pass
        if isiterable(ret):
            nc = np.zeros((len(self), len(ret)))
        else:
            nc = np.zeros(len(self))
        # Evaluate the data row by row
        for ix, r in enumerate(self.rows()):
            ret = func(r, **kargs)
            if isiterable(ret) and not isinstance(ret, np.ndarray):
                ret = np.ma.MaskedArray(ret)
            nc[ix] = ret
        # Work out how to handle the result
        if nc.ndim == 1:
            if header is None:
                header = func.__name__
            self.add_column(nc, header=header, index=col, replace=replace, setas=self.setas[col])
            ret = self
        else:
            if not replace:
                ret = self.clone
            else:
                ret = self
            ret.data = nc
            if header is not None:
                ret.column_headers = header
        return ret

    def clip(self, clipper, column=None):
        """Clips the data based on the column and the clipper value.

        Args:
            column (index):
                Column to look for the maximum in
            clipper (tuple or array):
                Either a tuple of (min,max) or a numpy.ndarray -
                in which case the max and min values in that array will be
                used as the clip limits
        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        Note:
            If column is not defined (or is None) the :py:attr:`DataFile.setas` column
            assignments are used.
        """
        if column is None:
            col = self.setas._get_cols("ycol")
        else:
            col = self.find_col(column)
        clipper = (min(clipper), max(clipper))
        return self.del_rows(col, lambda x, y: x < clipper[0] or x > clipper[1])

    def decompose(self, xcol=None, ycol=None, sym=None, asym=None, replace=True, **kwords):
        """Given (x,y) data, decomposes the y part into symmetric and antisymmetric contributions in x.

        Keyword Arguments:
            xcol (index):
                Index of column with x data - defaults to first x column in self.setas
            ycol (index or list of indices):
                indices of y column(s) data
            sym (index):
                Index of column to place symmetric data in default, append to end of data
            asym (index):
                Index of column for asymmetric part of ata. Defaults to appending to end of data
            replace (bool):
                Overwrite data with output (true)

        Returns:
            self: The newly modified :py:class:`AnalysisMixin`.

        Example:
            .. plot:: samples/decompose.py
                :include-source:
                :outname: decompose
        """
        if xcol is None and ycol is None:
            if "_startx" in kwords:
                startx = kwords["_startx"]
                del kwords["_startx"]
            else:
                startx = 0
            cols = self.setas._get_cols(startx=startx)
            xcol = cols["xcol"]
            ycol = cols["ycol"]
        xcol = self.find_col(xcol)
        ycol = self.find_col(ycol)
        if isinstance(ycol, list):
            ycol = ycol[0]  # FIXME should work with multiple output columns
        if np.abs((self // xcol).min()) < np.abs((self // xcol).max()):
            pxdata = self.search(xcol, lambda x, r: x < 0, xcol)
        else:
            pxdata = self.search(xcol, lambda x, r: x > 0, xcol)
        xdata = np.sort(np.append(-pxdata, pxdata))
        self.data = self.interpolate(xdata, xcol=xcol)
        ydata = self.data[:, ycol]
        symd = (ydata + ydata[::-1]) / 2
        asymd = (ydata - ydata[::-1]) / 2
        if sym is None:
            self &= symd
            self.column_headers[-1] = "Symmetric Data"
        else:
            self.add_column(symd, header="Symmetric Data", index=sym, replace=replace)
        if asym is None:
            self &= asymd
            self.column_headers[-1] = "Asymmetric Data"
        else:
            self.add_column(asymd, header="Symmetric Data", index=asym, replace=replace)

        return self

    def integrate(
        self,
        xcol=None,
        ycol=None,
        result=None,
        header=None,
        result_name=None,
        output="data",
        bounds=lambda x, y: True,
        **kargs,
    ):
        """Integrate a column of data, optionally returning the cumulative integral.

        Args:
            xcol (index):
                The X data column index (or header)
            ycol (index)
            The Y data column index (or header)

        Keyword Arguments:
            result (index or None):
                Either a column index (or header) to overwrite with the cumulative data,
                or True to add a new column or None to not store the cumulative result.
            result_name (str):
                The metadata name for the final result
            header (str):
                The name of the header for the results column.
            output (Str):
                What to return - 'data' (default) - this object, 'result': final result
            bounds (callable):
                A function that evaluates for each row to determine if the data should be integrated over.
            **kargs:
                Other keyword arguments are fed direct to the scipy.integrate.cumtrapz method

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        Note:
            This is a pass through to the scipy.integrate.cumtrapz routine which just uses trapezoidal integration.
            A better alternative would be to offer a variety of methods including simpson's rule and interpolation
            of data. If xcol or ycol are not specified then the current values from the
            :py:attr:`Stoner.Core.DataFile.setas` attribute are used.
        """
        _ = self._col_args(xcol=xcol, ycol=ycol)

        working = self.search(_.xcol, bounds)
        working = ma.mask_rowcols(working, axis=0)
        xdat = working[:, self.find_col(_.xcol)]
        ydat = working[:, self.find_col(_.ycol)]
        ydat = np.atleast_2d(ydat).T

        final = []
        for i in range(ydat.shape[1]):
            yd = ydat[:, i]
            resultdata = cumtrapz(yd, xdat, **kargs)
            resultdata = np.append(np.array([0]), resultdata)
            if result is not None:
                header = header if header is not None else f"Integral of {self.column_headers[_.ycol]}"
                if isinstance(result, bool) and result:
                    self.add_column(resultdata, header=header, replace=False)
                else:
                    result_name = self.column_headers[self.find_col(result)]
                    self.add_column(resultdata, header=header, index=result, replace=(i == 0))
            final.append(resultdata[-1])
        if len(final) == 1:
            final = final[0]
        else:
            final = np.array(final)
        result_name = result_name if result_name is not None else header if header is not None else "Area"
        self[result_name] = final
        if output.lower() == "result":
            return final
        return self

    def normalise(self, target=None, base=None, replace=True, header=None, scale=None, limits=(0.0, 1.0)):
        """Normalise data columns by dividing through by a base column value.

        Args:
            target (index):
                One or more target columns to normalise can be a string, integer or list of strings or integers.
                If None then the default 'y' column is used.

        Keyword Arguments:
            base (index):
                The column to normalise to, can be an integer or string. **Deprecated** can also be a tuple (low,
                high) being the output range
            replace (bool):
                Set True(default) to overwrite  the target data columns
            header (string or None):
                The new column header - default is target name(norm)
            scale (None or tuple of float,float):
                Output range after normalising - low,high or None to map to -1,1
            limits (float,float):
                (low,high) - Take the input range from the *high* and *low* fraction of the input when sorted.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        Notes:
            The *limits* parameter is used to set the input scale being normalised from - if the data has a few
            outliers then this setting can be used to clip the input range before normalising. The parameters in
            the limit are the values at the *low* and *high* fractions of the cumulative distribution function of
            the data.
        """
        _ = self._col_args(scalar=True, ycol=target)

        target = _.ycol
        if not isinstance(target, list):
            target = [self.find_col(target)]
        for t in target:
            if header is None:
                header = self.column_headers[self.find_col(t)] + "(norm)"
            else:
                header = str(header)
            if not isTuple(base, float, float) and base is not None:
                self.divide(t, base, header=header, replace=replace)
            else:
                if isTuple(base, float, float):
                    scale = base
                elif scale is None:
                    scale = (-1.0, 1.0)
                if not isTuple(scale, float, float):
                    raise ValueError("limit parameter is either None, or limit or base is a tuple of two floats.")
                data = self.column(t).ravel()
                data = np.sort(data[~np.isnan(data)])
                if limits != (0.0, 1.0):
                    low, high = limits
                    low = data[int(low * data.size)]
                    high = data[int(high * data.size)]
                else:
                    high = data.max()
                    low = data.min()
                data = np.copy(self.data[:, t])
                data = np.where(data > high, high, np.where(data < low, low, data))
                scl, sch = scale
                data = (data - low) / (high - low) * (sch - scl) + scl
                setas = self.setas.clone
                self.add_column(data, index=t, replace=replace, header=header)
                self.setas = setas
        return self

    def stitch(self, other, xcol=None, ycol=None, overlap=None, min_overlap=0.0, mode="All", func=None, p0=None):
        r"""Apply a scaling to this data set to make it stich to another dataset.

        Args:
            other (DataFile):
                Another data set that is used as the base to stitch this one on to
            xcol,ycol (index or None):
                The x and y data columns. If left as None then the current setas attribute is used.

        Keyword Arguments:
            overlap (tuple of (lower,higher) or None):
                The band of x values that are used in both data sets to match,
                if left as None, thenthe common overlap of the x data is used.
            min_overlap (float):
                If you know that overlap must be bigger than a certain amount, the bounds between the two
                data sets needs to be adjusted. In this case min_overlap shifts the boundary of the overlap
                on this DataFile.
            mode (str):
                Unless *func* is specified, controls which parameters are actually variable, defaults to all of them.
            func (callable):
                a stitching function that transforms :math:`(x,y)\rightarrow(x',y')`. Default is to use
                functions defined by *mode*
            p0 (iterable):
                if func is not None then p0 should be the starting values for the stitching function parameters

        Returns:
            (:py:class:`Stoner.Data`):
                A copy of the current :py:class:`AnalysisMixin` with the x and y data columns adjusted to stitch

        To stitch the data together, the x and y data in the current data file is transforms so that
        :math:`x'=x+A` and :math:`y'=By+C` where :math:`A,B,C` are constants and :math:`(x',y')` are close matches
        to the :math:`(x,y)` data in *other*. The algorithm assumes that the overlap region contains equal
        numbers of :math:`(x,y)` points *mode* controls whether A,B, and C are fixed or adjustable

            - "All" - all three parameters adjustable
            - "Scale y, shift x" - C is fixed at 0.0
            - "Scale and shift y" A is fixed at 0.0
            - "Scale y" - only B is adjustable
            - "Shift y" - Only c is adjsutable
            - "Shift x" - Only A is adjustable
            - "Shift both" - B is fixed at 1.0

        See Also:
            User Guide section :ref:`stitch_guide`

        Example:
            .. plot:: samples/stitch-int-overlap.py
                :include-source:
                :outname:  stitch_int_overlap
        """
        _ = self._col_args(xcol=xcol, ycol=ycol, scalar=True)
        points = self.column([_.xcol, _.ycol])
        points = points[points[:, 0].argsort(), :]
        points[:, 0] += min_overlap
        otherpoints = other.column([_.xcol, _.ycol])
        otherpoints = otherpoints[otherpoints[:, 0].argsort(), :]
        self_second = np.max(points[:, 0]) > np.max(otherpoints[:, 0])
        if overlap is None:  # Calculate the overlap
            lower = max(np.min(points[:, 0]), np.min(otherpoints[:, 0]))
            upper = min(np.max(points[:, 0]), np.max(otherpoints[:, 0]))
        elif isinstance(overlap, int) and overlap > 0:
            if self_second:
                lower = points[0, 0]
                upper = points[overlap, 0]
            else:
                lower = points[-overlap - 1, 0]
                upper = points[-1, 0]
        elif (
            isinstance(overlap, tuple)
            and len(overlap) == 2
            and isinstance(overlap[0], float and isinstance(overlap[1], float))
        ):
            lower = min(overlap)
            upper = max(overlap)
        inrange = np.logical_and(points[:, 0] >= lower, points[:, 0] <= upper)
        points = points[inrange]
        num_pts = points.shape[0]
        if self_second:
            otherpoints = otherpoints[-num_pts - 1 : -1]
        else:
            otherpoints = otherpoints[0:num_pts]
        x = points[:, 0]
        y = points[:, 1]
        xp = otherpoints[:, 0]
        yp = otherpoints[:, 1]
        if func is None:
            opts = {
                "all": (lambda x, y, A, B, C: (x + A, y * B + C)),
                "scale y and shift x": (lambda x, y, A, B: (x + A, B * y)),
                "scale and shift y": (lambda x, y, B, C: (x, y * B + C)),
                "scale y": (lambda x, y, B: (x, y * B)),
                "shift y": (lambda x, y, C: (x, y + C)),
                "shift both": (lambda x, y, A, C: (x + A, y + C)),
            }
            defaults = {
                "all": [1, 2, 3],
                "scale y,shift x": [1, 2],
                "scale and shift y": [2, 3],
                "scale y": [2],
                "shift y": [3],
                "shift both": [1, 3],
            }
            A0 = np.mean(xp) - np.mean(x)
            C0 = np.mean(yp) - np.mean(y)
            B0 = (np.max(yp) - np.min(yp)) / (np.max(y) - np.min(y))
            p = np.array([0, A0, B0, C0])
            assertion(isinstance(mode, string_types), "mode keyword should be a string if func is not defined")
            mode = mode.lower()
            assertion(mode in opts, f"mode keyword should be one of {opts.keys}")
            func = opts[mode]
            p0 = p[defaults[mode]]
        else:
            assertion(callable(func), "Keyword func should be callable if given")
            args = getfullargspec(func)[0]  # pylint: disable=W1505
            assertion(isiterable(p0), "Keyword parameter p0 shoiuld be iterable if keyword func is given")
            assertion(
                len(p0) == len(args) - 2, "Keyword p0 should be the same length as the optional arguments to func"
            )
        # This is a bit of a hack, we turn (x,y) points into a 1D array of x and then y data
        set1 = np.append(x, y)
        set2 = np.append(xp, yp)
        assertion(len(set1) == len(set2), "The number of points in the overlap are different in the two data sets")

        def transform(set1, *p):
            """Construct the wrapper function to fit for transform."""
            m = int(len(set1) / 2)
            x = set1[:m]
            y = set1[m:]
            tmp = func(x, y, *p)
            out = np.append(tmp[0], tmp[1])
            return out

        popt, pcov = curve_fit(transform, set1, set2, p0=p0)  # Curve fit for optimal A,B,C
        perr = np.sqrt(np.diagonal(pcov))
        self.data[:, _.xcol], self.data[:, _.ycol] = func(self.data[:, _.xcol], self.data[:, _.ycol], *popt)
        self["Stitching Coefficients"] = list(popt)
        self["Stitching Coefficient Errors"] = list(perr)
        self["Stitching overlap"] = (lower, upper)
        self["Stitching Window"] = num_pts

        return self

    def threshold(self, threshold, **kargs):
        """Find partial indices where the data in column passes the threshold, rising or falling.

        Args:
            threshold (float):
                Value to look for in column col

        Keyword Arguments:
            col (index):
                Column index to look for data in
            rising (bool):
                look for case where the data is increasing in value (default True)
            falling (bool):
                look for case where data is fallinh in value (default False)
            xcol (index, bool or None):
                rather than returning a fractional row index, return the
                interpolated value in column xcol. If xcol is False, then return a complete row
                all_vals (bool): return all crossing points of the threshold or just the first. (default False)
            transpose (bbool):
                Swap the x and y columns around - this is most useful when the column assignments
                have been done via the setas attribute
            all_vals (bool):
                Return all values that match the criteria, or just the first in the file.

        Returns:
            (float):
                Either a sing;le fractional row index, or an in terpolated x value

        Note:
            If you don't specify a col value or set it to None, then the assigned columns via the
            :py:attr:`DataFile.setas` attribute will be used.

        Warning:
            There has been an API change. Versions prior to 0.1.9 placed the column before the threshold in the
            positional argument list. In order to support the use of assigned columns, this has been swapped to the
            present order.
        """
        DataArray = type(self.data)
        col = kargs.pop("col", None)
        xcol = kargs.pop("xcol", None)
        _ = self._col_args(xcol=xcol, ycol=col)

        col = _.ycol
        if xcol is None and _.has_xcol:
            xcol = _.xcol

        rising = kargs.pop("rising", True)
        falling = kargs.pop("falling", False)
        all_vals = kargs.pop("all_vals", False)

        current = self.column(col)

        # Recursively call if we've got an iterable threshold
        if isiterable(threshold):
            if isinstance(xcol, bool) and not xcol:
                ret = np.zeros((len(threshold), self.shape[1]))
            else:
                ret = np.zeros_like(threshold).view(type=DataArray)
            for ix, th in enumerate(threshold):
                ret[ix] = self.threshold(th, col=col, xcol=xcol, rising=rising, falling=falling, all_vals=all_vals)
            # Now we have to clean up the  retujrn list into a DataArray
            if isinstance(xcol, bool) and not xcol:  # if xcol was False we got a complete row back
                ch = self.column_headers
                ret.setas = self.setas.clone
                ret.column_headers = ch
                ret.i = ret[0].i
            else:  # Either xcol was None so we got indices or we got a specified column back
                if xcol is not None:  # Specific column
                    ret = np.atleast_2d(ret)
                    ret.column_headers = [self.column_headers[self.find_col(xcol)]]
                    ret.i = [r.i for r in ret]
                    ret.setas = "x"
                    ret.isrow = False
                else:
                    ret.column_headers = ["Index"]
                    ret.isrow = False
            return ret
        ret = _threshold(threshold, current, rising=rising, falling=falling)
        if not all_vals:
            ret = [ret[0]] if np.any(ret) else []

        if isinstance(xcol, bool) and not xcol:
            retval = self.interpolate(ret, xcol=False)
            retval.setas = self.setas.clone
            retval.setas.shape = retval.shape
            retval.i = ret
            ret = retval
        elif xcol is not None:
            retval = self.interpolate(ret, xcol=False)[:, self.find_col(xcol)]
            # if retval.ndim>0:   #not sure what this bit does but it's throwing errors for a simple threshold
            # retval.setas=self.setas.clone
            # retval.setas.shape=retval.shape
            # retval.i=ret
            ret = retval
        else:
            ret = DataArray(ret)
        if not all_vals:
            if ret.size == 1:
                pass
            elif ret.size > 1:
                ret = ret[0]
            else:
                ret = []
        if isinstance(ret, DataArray):
            ret.isrow = True
        return ret



# -*- coding: utf-8 -*-
"""Ensure a consistent namespace for the rest of the package irrespective of Python language version."""
__all__ = [
    "str2bytes",
    "bytes2str",
    "get_filedialog",
    "string_types",
    "path_types",
    "int_types",
    "index_types",
    "classproperty",
    "mpl_version",
    "_lmfit",
    "makedirs",
    "Hyperspy_ok",
    "hs",
    "hsload",
    "which",
    "commonpath",
    "_jit",
    "_dummy",
]

from sys import version_info as __vi__, modules
from os import walk, makedirs
from os.path import join, commonpath
import fnmatch
from shutil import which
from pathlib import PurePath
from packaging.version import parse as version_parse
from inspect import signature
import re

import numpy as np
import scipy as sp
import matplotlib

_lmfit = True
np_version = version_parse(np.__version__)
sp_version = version_parse(sp.__version__)
mpl_version = version_parse(matplotlib.__version__)

try:  # This only works in PY 3.11 onwards
    modules["sre_parse"] = re._parser
    modules["sre_constants"] = re._constants
    modules["sre_compile"] = re._compiler
except AttributeError:
    pass

try:
    import hyperspy as hs  # Workaround an issue in hs 1.5.2 conda packages

    try:
        hsload = hs.load
    except (RuntimeError, AttributeError):
        try:
            from hyperspy import api

            hsload = api.load
        except (ImportError, AttributeError) as err:
            raise ImportError("Panic over hyperspy") from err

    HuperSpyVersion = [int(x) for x in hs.__version__.split(".")]
    if HuperSpyVersion[0] <= 1 and HuperSpyVersion[1] <= 3:
        raise ImportError(f"Hyperspy should be version 1.4 or above. Actual version is {hs.__version__}")
    Hyperspy_ok = True
except ImportError:
    Hyperspy_ok = False
    hs = None
    hsload = None

if __vi__[1] < 7:
    from re import _pattern_type  # pylint: disable = E0611
else:
    from re import Pattern as _pattern_type  # pylint: disable = E0611


def get_func_params(func):
    """Get the parameters for a function."""
    sig = signature(func)
    ret = {}
    for i, k in enumerate(sig.parameters):
        if i == 0:
            continue
        ret[k] = sig.parameters[k]
    return list(ret.keys())


string_types = (str,)
int_types = (int,)
path_types = (str, PurePath)

# #### Monkey patch numpy for removed attributes as a compatibiliyu hack
if np_version.minor >= 20:
    np.float = float
    np.bool = np.bool_
    np.str = str
    np.bool8 = np.bool_
    np.int0 = np.intp


def str2bytes(data):
    """Encode a unicode string into UTF-8."""
    if isinstance(data, bytes):
        return data
    return bytes(str(data), "utf-8")


def bytes2str(data):
    """Decode byte string back to univcode."""
    if isinstance(data, bytes):
        return data.decode("utf-8", "ignore")
    return data


def get_filedialog(what="file", **opts):
    """Wrap around Tk file dialog to manage creating file dialogs in a cross platform way.

    Args:
        what (str): What sort of a dialog to create - options are 'file','directory','save','files'
        **opts (dict): Arguments to pass through to the underlying dialog function.

    Returns:
        A file name or directory or list of files.
    """
    from .tools.widgets import fileDialog

    funcs = {"file": "OpenFile", "directory": "SelectDirectory", "files": "OpenFiles", "save": "SaveFile"}
    if what not in funcs:
        raise RuntimeError(f"Unable to recognise required file dialog type:{what}")
    return fileDialog.openDialog(mode=funcs[what], **opts)


if np_version.minor >= 20:
    int_types += (int, np.int0, np.int8, np.int16, np.int32, np.int64)
else:
    int_types += (np.int, np.int0, np.int8, np.int16, np.int32, np.int64)

index_types = string_types + int_types + (_pattern_type,)


def listdir_recursive(dirname, glob=None):
    """Make a recursive file list with optional globbing."""
    for dp, _, fn in walk(dirname):
        for f in fn:
            ret = join(dp, f)
            if glob is not None:
                if not fnmatch.fnmatch(ret, glob):
                    continue
            yield ret


class ClassPropertyDescriptor:
    """Supports adding class properties."""

    def __init__(self, fget, fset=None):
        """Initialise the descriptor."""
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        """Implement descriptor getter."""
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()


def classproperty(func):
    """Define a property to be a class property and not an instance property."""
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


def _jit(func, *_, **__):
    """Null decorator function."""
    return func


class _dummy:
    """A class that does nothing so that float64 can be an instance of it safely."""

    def jit(self, func, *_, **__):  # pylint: disable=no-self-use
        """Null decorator function."""
        return func

    def __call__(self, *args, **kargs):
        """Handle jit lines that call the type."""
        return self.jit

    def __getitem__(self, *args, **kargs):
        """Hande jit calls to array types."""
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test fitting mixin classes. Much of the code is also tested by the doc/samples scripts - so this is just
filling in some gaps.

Created on Sat Aug 24 22:56:59 2019

@author: phygbu
"""
import unittest
import numpy as np


from Stoner import Data


class ColumnOps_test(unittest.TestCase):
    def setUp(self):
        """Create a test data set."""
        x = np.linspace(1, 10, 10)
        y = 2 * x - 3
        dy = np.abs(y / 100)
        z = x + 4
        dz = np.abs(z / 100)
        self.data = Data(
            np.column_stack((x, y, dy, z, dz)), column_headers=["Tine", "Signal 1", "d1", "Signal 2", "d2"]
        )

    def test_add(self):
        self.data.add(1, 3, header="Add")
        self.assertTrue(
            np.all(self.data // "Add" == (self.data // "Signal 1" + self.data // "Signal 2")),
            "Failed to add column correctly",
        )
        self.data.add((1, 2), (3, 4), header="Add", index="Add", replace=True)
        d_man = np.sqrt((self.data // "d1") ** 2 + (self.data // "d2") ** 2)
        self.assertTrue(np.allclose(self.data // -1, d_man), "Failed to calculate error in add")
        self.data.add(1, 3.0, index="Add", replace=True)
        self.assertTrue(np.all(self.data // 5 == self.data // 1 + 3), "Failed to add with a constant")
        self.data.add(1, np.ones(10), index=5, replace=True)
        self.assertTrue(np.all(self.data // 5 == self.data // 1 + 1), "Failed to add with a array")

    def test_subtract(self):
        self.data.subtract(1, 3, header="Subtract")
        self.assertTrue(
            np.all(self.data // "Subtract" == (self.data // "Signal 1" - self.data // "Signal 2")),
            "Failed to add column correctly",
        )
        self.data.subtract((1, 2), (3, 4), header="Subtract", index="Subtract", replace=True)
        d_man = np.sqrt((self.data // "d1") ** 2 + (self.data // "d2") ** 2)
        self.assertTrue(np.allclose(self.data // -1, d_man), "Failed to calculate error in add")
        self.data.subtract(1, 3.0, index="Subtract", replace=True)
        self.assertTrue(np.all(self.data // 5 == self.data // 1 - 3), "Failed to subtract with a constant")
        self.data.subtract(1, np.ones(10), index=5, replace=True)
        self.assertTrue(np.all(self.data // 5 == self.data // 1 - 1), "Failed to subtract with a array")

    def test_multiply(self):
        self.data.multiply(1, 3, header="Multiply")
        self.assertTrue(
            np.all((self.data // "Multiply") == ((self.data // "Signal 1") * (self.data // "Signal 2"))),
            "Failed to add column correctly",
        )
        self.data.multiply((1, 2), (3, 4), header="Multiply", index="Multiply", replace=True)
        d_man = np.sqrt(2) * 0.01 * np.abs(self.data // -2)
        self.assertTrue(np.allclose(self.data // -1, d_man), "Failed to calculate error in add")
        self.data.multiply(1, 3.0, index="Multiply", replace=True)
        self.assertTrue(np.all(self.data // 5 == (self.data // 1) * 3), "Failed to multiply with a constant")
        self.data.multiply(1, 2 * np.ones(10), index=5, replace=True)
        self.assertTrue(np.all(self.data // 5 == (self.data // 1) * 2), "Failed to multiply with a array")

    def test_divide(self):
        self.data.divide(1, 3, header="Divide")
        self.assertTrue(
            np.all((self.data // "Divide") == ((self.data // "Signal 1") / (self.data // "Signal 2"))),
            "Failed to add column correctly",
        )
        self.data.divide((1, 2), (3, 4), header="Divide", index="Divide", replace=True)
        d_man = np.sqrt(2) * 0.01 * np.abs(self.data // -2)
        self.assertTrue(np.allclose(self.data // -1, d_man), "Failed to calculate error in diffsum")
        self.data.divide(1, 3.0, index="Divide", replace=True)
        self.assertTrue(np.all(self.data // 5 == (self.data // 1) / 3), "Failed to add with a constant")
        self.data.divide(1, 2 * np.ones(10), index=5, replace=True)
        self.assertTrue(np.all(self.data // 5 == (self.data // 1) / 2), "Failed to add with a array")

    def test_diffsum(self):
        self.data.diffsum(1, 3, header="Diffsum")
        a = self.data // 1
        b = self.data // 3
        man = (a - b) / (a + b)
        self.assertTrue(np.all((self.data // "Diffsum") == man), "Failed to diffsum column correctly")
        self.data.diffsum((1, 2), (3, 4), header="Diffsum", index="Diffsum", replace=True)

    def test_limits(self):
        self.data.setas = "x..ye"
        self.assertEqual(self.data.min(1), (-1, 0), "Minimum method failed")
        self.assertEqual(self.data.min(), (5.0, 0), "Minimum method failed")
        self.assertEqual(self.data.min(1, bounds=lambda r: r[0] > 2), (3.0, 2.0), "Max with bounds failed")
        self.assertEqual(self.data.max(), (14, 9), "Max method failed")
        self.assertEqual(self.data.max(1), (17, 9), "Max method failed")
        self.assertEqual(self.data.max(1, bounds=lambda r: r[0] < 5), (5.0, 3.0), "Max with bounds failed")
        self.assertEqual(self.data.span(), (5.0, 14.0), "span method failed")
        self.assertEqual(self.data.span(1), (-1.0, 17.0), "span method failed")
        self.assertEqual(self.data.span(1, bounds=lambda r: 2 < r.i < 8), (5.0, 13.0), "span with bounds failed")

    def test_stats(self):
        self.assertEqual(self.data.mean(1), 8.0, "Simple channel mean failed")
        self.assertTrue(
            np.allclose(self.data.mean(1, sigma=2), (0.048990998729652346, 0.031144823004794875)),
            "Channel mean with sigma failed",
        )
        self.assertAlmostEqual(self.data.std(1), 6.0553007081949835, msg="Simple Standard Deviation failed")
        self.assertAlmostEqual(self.data.std(1, 2), 2.7067331877422456, msg="Simple Standard Deviation failed")


if __name__ == "__main__":  # Run some tests manually to allow debugging
    test = ColumnOps_test("test_add")
    test.setUp()
    unittest.main()
    # test.test_add()
    # test.setUp()
    # test.test_subtract()
    # test.setUp()
    # test.test_multiply()



"""Demo of new Stoner.Analysis.AnalyseFile.lmfit."""

# pylint: disable=invalid-name
from os.path import join

from Stoner.analysis.fitting.models import cfg_data_from_ini, cfg_model_from_ini
from Stoner import __home__

config = join(__home__, "..", "scripts", "PCAR-New.ini")
datafile = join(__home__, "..", "sample-data", "PCAR Co Data.csv")

d = cfg_data_from_ini(config, datafile)
model, p0 = cfg_model_from_ini(config, data=d)

d.x += 0.25
d.setas = "xy"
d.plot(fmt="r.")  # plot the data
fit = d.lmfit(model, result=True, header="lmfit")
fit = d.odr(model, result=True, header="odr", prefix="odr")
fit = d.differential_evolution(model, result=True, header="odr", prefix="diff")
d.setas = "x.yyy"
d.plot(fmt=["b-", "g-", "m-"], label=["lmfit", "odr", "diff.ev."])

d.annotate_fit(
    model, x=0.05, y=0.75, fontdict={"size": "x-small", "color": "blue"}
)
d.annotate_fit(
    model,
    x=0.05,
    y=0.5,
    fontdict={"size": "x-small", "color": "green"},
    prefix="odr",
)
d.annotate_fit(
    model,
    x=0.05,
    y=0.25,
    fontdict={"size": "x-small", "color": "magenta"},
    prefix="diff",
)



"""Example of nDimArrhenius Fit."""

# pylint: disable=invalid-name
from numpy import linspace
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.thermal import modArrhenius, ModArrhenius

# Make some data
T = linspace(200, 350, 101)
R = modArrhenius(T, 1e6, 0.5, 1.5) * normal(scale=0.00005, loc=1.0, size=len(T))
d = Data(T, R, setas="xy", column_headers=["T", "Rate"])

# Curve_fit on its own
d.curve_fit(modArrhenius, p0=[1e6, 0.5, 1.5], result=True, header="curve_fit")
d.setas = "xyy"
d.plot(fmt=["r.", "b-"])
d.annotate_fit(modArrhenius, x=0.2, y=0.5, mode="eng")

# lmfit using lmfit guesses
fit = ModArrhenius()
p0 = [1e6, 0.5, 1.5]
d.lmfit(fit, result=True, header="lmfit")
d.setas = "x..y"
d.plot()
d.annotate_fit(ModArrhenius, x=0.2, y=0.25, prefix="ModArrhenius", mode="eng")

d.title = "Modified Arrhenius Test Fit"
d.ylabel = "Rate"
d.xlabel = "Temperature (K)"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Implements DataFile like classes for various large scale facilities."""

# Standard Library imports
import linecache
import re

import numpy as np

from ...compat import str2bytes
from ...core.base import string_to_type
from ...tools.file import FileManager
from ...core.exceptions import StonerLoadError
from ..decorators import register_loader

try:
    import fabio
except ImportError:
    fabio = None


def _bnl_find_lines(new_data):
    """Return an array of ints [header_line,data_line,scan_line,date_line,motor_line]."""
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as fp:
        new_data.line_numbers = [0, 0, 0, 0, 0]
        counter = 0
        for line in fp:
            counter += 1
            if counter == 1 and line[0] != "#":
                raise StonerLoadError("Not a BNL File ?")
            if len(line) < 2:
                continue  # if there's nothing written on the line go to the next
            if line[0:2] == "#L":
                new_data.line_numbers[0] = counter
            elif line[0:2] == "#S":
                new_data.line_numbers[2] = counter
            elif line[0:2] == "#D":
                new_data.line_numbers[3] = counter
            elif line[0:2] == "#P":
                new_data.line_numbers[4] = counter
            elif line[0] in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                new_data.line_numbers[1] = counter
                break


def _bnl_get_metadata(new_data):
    """Load metadta from file.

    Metadata found is scan number 'Snumber', scan type and parameters 'Stype',
    scan date/time 'Sdatetime' and z motor position 'Smotor'.
    """
    scanLine = linecache.getline(new_data.filename, new_data.line_numbers[2])
    dateLine = linecache.getline(new_data.filename, new_data.line_numbers[3])
    motorLine = linecache.getline(new_data.filename, new_data.line_numbers[4])
    new_data.__setitem__("Snumber", scanLine.split()[1])
    tmp = "".join(scanLine.split()[2:])
    new_data.__setitem__("Stype", "".join(tmp.split(",")))  # get rid of commas
    new_data.__setitem__("Sdatetime", dateLine[3:-1])  # don't want \n at end of line so use -1
    new_data.__setitem__("Smotor", motorLine.split()[3])


def _parse_bnl_data(new_data):
    """Parse BNL data.

     The meta data is labelled by #L type tags
    so easy to find but #L must be excluded from the result.
    """
    _bnl_find_lines(new_data)
    # creates a list, line_numbers, formatted [header_line,data_line,scan_line,date_line,motor_line]
    header_string = linecache.getline(new_data.filename, new_data.line_numbers[0])
    header_string = re.sub(r'["\n]', "", header_string)  # get rid of new line character
    header_string = re.sub(r"#L", "", header_string)  # get rid of line indicator character
    column_headers = map(lambda x: x.strip(), header_string.split())
    _bnl_get_metadata(new_data)
    try:
        new_data.data = np.genfromtxt(new_data.filename, skip_header=new_data.line_numbers[1] - 1)
    except IOError:
        new_data.data = np.array([0])
        print(f"Did not import any data for {new_data.filename}")
    new_data.column_headers = column_headers


@register_loader(patterns=(".txt", 64), mime_types=("text/plain", 64), name="BNLFile", what="Data")
def load_bnl(new_data, filename, *args, **kargs):  # pylint: disable=unused-argument
    """Load the file from disc.

    Args:
        filename (string or bool):
            File to load. If None then the existing filename is used, if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.

    Notes:
        Overwrites load method in Core.DataFile class, no header positions and data
        positions are needed because of the hash title structure used in BNL files.

        Normally its good to use _parse_plain_data method from Core.DataFile class
        to load data but unfortunately Brookhaven data isn't very plain so there's
        a new method below.
    """
    new_data.filename = filename
    try:
        _parse_bnl_data(new_data)  # call an internal function rather than put it in load function
    except (IndexError, TypeError, ValueError, StonerLoadError) as err:
        raise StonerLoadError("Not parseable as a NFLS file!") from err
    linecache.clearcache()
    return new_data


def _read_mdaascii_header(data, new_data, i):
    """Read the header block."""
    for i[0], line in enumerate(data):
        line.strip()
        if "=" in line:
            parts = line[2:].split("=")
            new_data[parts[0].strip()] = string_to_type("".join(parts[1:]).strip())
        elif line.startswith("#  Extra PV:"):
            # Onto the next metadata bit
            break


def _read_mdaascii_metadata(data, new_data, i):
    """Read the metadata block."""
    pvpat = re.compile(r"^#\s+Extra\s+PV\s\d+\:(.*)")
    for i[1], line in enumerate(data):
        if line.strip() == "":
            continue
        if line.startswith("# Extra PV"):
            res = pvpat.match(line)
            bits = [b.strip().strip(r'"') for b in res.group(1).split(",")]
            if bits[1] == "":
                key = bits[0]
            else:
                key = bits[1]
            if len(bits) > 3:
                key = f"{key} ({bits[3]})"
            new_data[key] = string_to_type(bits[2])
        else:
            break  # End of Extra PV stuff
    else:
        raise StonerLoadError("Overran Extra PV Block")
    for i[2], line in enumerate(data):
        line.strip()
        if line.strip() == "":
            continue
        elif line.startswith("# Column Descriptions:"):
            break  # Start of column headers now
        elif "=" in line:
            parts = line[2:].split("=")
            new_data[parts[0].strip()] = string_to_type("".join(parts[1:]).strip())
    else:
        raise StonerLoadError("Overran end of scan header before column descriptions")


def _read_mdaascii_columns(data, new_data, i):
    """Reads the column header block."""
    colpat = re.compile(r"#\s+\d+\s+\[([^\]]*)\](.*)")
    column_headers = []
    for i[3], line in enumerate(data):
        res = colpat.match(line)
        line.strip()
        if line.strip() == "":
            continue
        elif line.startswith("# 1-D Scan Values"):
            break  # Start of data
        elif res is not None:
            if "," in res.group(2):
                bits = [b.strip() for b in res.group(2).split(",")]
                if bits[-2] == "":
                    colname = bits[0]
                else:
                    colname = bits[-2]
                if bits[-1] != "":
                    colname += f"({bits[-1]})"
                if colname in column_headers:
                    colname = f"{bits[0]}:{colname}"
            else:
                colname = res.group(1).strip()
            column_headers.append(colname)
    else:
        raise StonerLoadError("Overand the end of file without reading data")
    new_data.column_headers = column_headers


@register_loader(patterns=(".txt", 32), mime_types=("text/plain", 32), name="MDAASCIIFile", what="Data")
def load_mdaasci(new_data, filename, *args, **kargs):  # pylint: disable=unused-argument
    """Load function. File format has space delimited columns from row 3 onwards."""
    new_data.filename = filename
    i = [0, 0, 0, 0]
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as data:  # Slightly ugly text handling
        if next(data).strip() != "## mda2ascii 1.2 generated output":
            raise StonerLoadError("Not a file mda2ascii")
        _read_mdaascii_header(data, new_data, i)
        _read_mdaascii_metadata(data, new_data, i)
        _read_mdaascii_columns(data, new_data, i)
        new_data.data = np.genfromtxt(data)  # so that's ok then !
    return new_data


@register_loader(patterns=(".dat", 16), mime_types=("text/plain", 16), name="OpenGDAFile", what="Data")
def load_gda(new_data, filename=None, *args, **kargs):
    """Load an OpenGDA file.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    new_data.filename = filename
    i = 0
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as f:
        for i, line in enumerate(f):
            line = line.strip()
            if i == 0 and line != "&SRS":
                raise StonerLoadError(f"Not a GDA File from Rasor ?\n{line}")
            if "&END" in line:
                break
            parts = line.split("=")
            if len(parts) != 2:
                continue
            key = parts[0]
            value = parts[1].strip()
            new_data.metadata[key] = string_to_type(value)
        column_headers = f.readline().strip().split("\t")
        new_data.data = np.genfromtxt([str2bytes(l) for l in f], dtype="float", invalid_raise=False)
    new_data.column_headers = column_headers
    return new_data


@register_loader(patterns=(".dat", 16), mime_types=("text/plain", 16), name="SNSFile", what="Data")
def load_sns(new_data, filename=None, *args, **kargs):
    """Load function. File format has space delimited columns from row 3 onwards."""
    new_data.filename = filename
    with FileManager(new_data.filename, "r", errors="ignore", encoding="utf-8") as data:  # Slightly ugly text handling
        line = data.readline()
        if not line.strip().startswith(
            "# Datafile created by QuickNXS 0.9.39"
        ):  # bug out oif we don't like the header
            raise StonerLoadError("Not a file from the SNS BL4A line")
        for line in data:
            if line.startswith("# "):  # We're in the header
                line = line[2:].strip()  # strip the header and whitespace

            if line.startswith("["):  # Look for a section header
                section = line.strip().strip("[]")
                if section == "Data":  # The Data section has one line of column headers and then data
                    header = next(data)[2:].split("\t")
                    column_headers = [h.strip() for h in header]
                    new_data.data = np.genfromtxt(data)  # we end by reading the raw data
                elif section == "Global Options":  # This section can go into metadata
                    for line in data:
                        line = line[2:].strip()
                        if line.strip() == "":
                            break
                        else:
                            new_data[line[2:10].strip()] = line[11:].strip()
                elif (
                    section == "Direct Beam Runs" or section == "Data Runs"
                ):  # These are constructed into lists ofg dictionaries for each file
                    sec = list()
                    header = next(data)
                    header = header[2:].strip()
                    keys = [s.strip() for s in header.split("  ") if s.strip()]
                    for line in data:
                        line = line[2:].strip()
                        if line == "":
                            break
                        else:
                            values = [s.strip() for s in line.split("  ") if s.strip()]
                            sec.append(dict(zip(keys, values)))
                    new_data[section] = sec
            else:  # We must still be in the opening un-labelled section of meta data
                if ":" in line:
                    i = line.index(":")
                    key = line[:i].strip()
                    value = line[i + 1 :].strip()
                    new_data[key.strip()] = value.strip()
    new_data.column_headers = column_headers
    return new_data


if fabio:

    @register_loader(
        patterns=(".edf", 16),
        mime_types=[("application/octet-stream", 16), ("text/plain", 15)],
        name="ESRF_DataFile",
        what="Data",
    )
    def load_esrf(self, filename=None, *args, **kargs):
        """Load function. File format has space delimited columns from row 3 onwards."""
        if filename is None or not filename:
            self.get_filename("r")
        else:
            self.filename = filename
        try:
            img = fabio.edfimage.edfimage().read(self.filename)
            self.data = img.data
            self.metadata.update(img.header)
            return self
        except (OSError, ValueError, TypeError, IndexError) as err:
            raise StonerLoadError("Not an ESRF data file !") from err



"""Add an inset to a plot."""

# pylint: disable=invalid-name
from Stoner import Data

p = Data("sample.txt", setas="xy")
p.plot()
p.inset(loc=1, width="50%", height="50%")
p.setas = "x.y"
p.plot()
p.title = ""  # Turn off the inset title
